        BAMS TECHNOLOGY LICENSE

The following license terms and conditions apply to this BAMS
software ("Software"), unless a different license is obtained from
Sendmail, Inc. at 6425 Christie Ave, Fourth Floor, Emeryville, CA 94608

=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

Copyright (c) 1999 - 2001
Sendmail, Inc.  All rights reserved.

License Terms:

Use, Modification and Redistribution (including distribution of
any modified or derived work) of the Software in source and binary
forms is permitted only if each of the following conditions is met:

1.  Redistributions of the Software and any modifications thereto
    or accompanying software that uses the Software are made at no
    charge beyond the reasonable cost of materials and delivery.

2.  Redistributions are accompanied by a complete copy of the Source
    Code or by an irrevocable offer to provide a complete copy of
    the Source Code for up to three years at the cost of materials
    and delivery.  Such redistributions must allow further use,
    modification, and redistribution of the Source Code under
    substantially the same terms as this license.  For the purposes
    of redistribution "Source Code" means the complete compilable
    and linkable source code of Software including all modifications
    and all accompanying software that uses the Software.  It does
    not include accompanying software modules or files that either
    (a) typically accompany the major components of the operating
    system on which the executable file runs, or (b) are distributed
    under another open source license, as defined by the Open Source
    Initiative (http://www.opensource.org).

3.  Redistributions of source code must retain the copyright notices
    as they appear in each source code file, these license terms,
    and the disclaimer/limitation of liability set forth as paragraph
    6 below.

4.  Redistributions in binary form must reproduce the Copyright
    Notice, these license terms, and the disclaimer/limitation of
    liability set forth as paragraph 6 below, in the documentation
    and/or other materials provided with the distribution.  For
    the purposes of binary distribution the "Copyright Notice"
    refers to the following language: "Copyright (c) 1999-2001
    Sendmail, Inc.  All rights reserved."

5.  Neither the name of Sendmail, Inc. nor its contributors may be
    used to endorse or promote products derived from this Software
    without specific prior written permission.  The name "sendmail"
    is a registered trademark of Sendmail, Inc.

6.  Disclaimer/Limitation of Liability: THIS SOFTWARE IS PROVIDED
    BY SENDMAIL, INC. AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
    PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL SENDMAIL, INC. OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
    STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
    ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
    ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.



#include "db.h"				/* berkeley DB database API */

#include <pthread.h>
#include <sql.h>
#include <sqlext.h>

#define MSDB_DATA_DIR "data"		/* database data subdirectory */
#define ODBC_MAXLEN_CNAME       16	/* Max len for cursor name, not really used */

/* : DOC - ODBC database environment context.  Contains all handles etc.
   : required to speak ODBC.  XXXYYYXXX: this is not yet thread-friendly.  */
typedef struct {
    pthread_t           mytid;          /* My pthread id  XXXYYYZZZ debugging */
    SQLHENV             henv;           /* ODBC environment handle */
    SQLHDBC             hdbc;           /* ODBC connection handle */
    char		autoc_p;	/* Flag: Autocommit mode on? */
    char                txn_p;          /* Flag: transaction in progress?  */
    char                ctxn_p;         /* Flag: child transaction in progress? 
*/
    /*
    ** Because we must carry our bound-to-assorted-columns buffers
    ** along with us.....
    */
    char                *in_idbuf;
    int                 *in_idbuflen;
    char                *res_idbuf;
    int                 *res_idbuflen;
    int                 *part;
    int                 *partmax;
    char                *res_valbuf;
    int                 *res_valbuflen;
} ODBC_ENV_t;

typedef struct {
    ODBC_ENV_t		*odbc_env;	/* ODBC environment container */
    char *              ctable;         /* Table name for cursor use */
    SQLHSTMT            cstmt;          /* ODBC statement handle for cursor ops */
    char                ccurname[ODBC_MAXLEN_CNAME+1];  /* Cursor name used */
    /*
    ** Because we must carry our bound-to-assorted-columns buffers
    ** along with us.....
    */
    char                *in_idbuf;
    int                 *in_idbuflen;
    char                *res_idbuf;
    int                 *res_idbuflen;
    int                 *part;
    int                 *partmax;
    char                *res_valbuf;
    int                 *res_valbuflen;
} ODBC_CURSOR_t;

/* : DOC - open database context.   This structure maintains open contexts for all
   : the database files..   We want to open and close
   : the contexts for the various databases exactly once per process because opening
   : and closing is an expensive operation. */
typedef struct {
    DB *db_handles[MS_DB_NUM_TYPES];    /* open database handles */
    DB_ENV * db_env;			/* database environment, but will
					   really be an (ODBC_ENV_t *) type. */
    MSDB_ERROR_CODE_t last_error;	/* XXX Nobody is using this! -SLF */
} MS_DB_CONTEXT_t;

/* : DOC - unique identifier type is a 64bit unsigned value */
typedef EC_UINT64                       MSDB_UID_t;

/* : DOC - concatenated identifiers are 128bit unsigned values */
typedef EC_BYTE *                       MSDB_UID_CONCAT_t;
#define MSDB_UID_CONCAT_t_SIZE          (2 * sizeof (MSDB_UID_t))

ODBC_ENV_t *get_per_thread_odbc_env(void);

/* ZZZ missing */
