        BAMS TECHNOLOGY LICENSE

The following license terms and conditions apply to this BAMS
software ("Software"), unless a different license is obtained from
Sendmail, Inc. at 6425 Christie Ave, Fourth Floor, Emeryville, CA 94608

=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

Copyright (c) 1999 - 2001
Sendmail, Inc.  All rights reserved.

License Terms:

Use, Modification and Redistribution (including distribution of
any modified or derived work) of the Software in source and binary
forms is permitted only if each of the following conditions is met:

1.  Redistributions of the Software and any modifications thereto
    or accompanying software that uses the Software are made at no
    charge beyond the reasonable cost of materials and delivery.

2.  Redistributions are accompanied by a complete copy of the Source
    Code or by an irrevocable offer to provide a complete copy of
    the Source Code for up to three years at the cost of materials
    and delivery.  Such redistributions must allow further use,
    modification, and redistribution of the Source Code under
    substantially the same terms as this license.  For the purposes
    of redistribution "Source Code" means the complete compilable
    and linkable source code of Software including all modifications
    and all accompanying software that uses the Software.  It does
    not include accompanying software modules or files that either
    (a) typically accompany the major components of the operating
    system on which the executable file runs, or (b) are distributed
    under another open source license, as defined by the Open Source
    Initiative (http://www.opensource.org).

3.  Redistributions of source code must retain the copyright notices
    as they appear in each source code file, these license terms,
    and the disclaimer/limitation of liability set forth as paragraph
    6 below.

4.  Redistributions in binary form must reproduce the Copyright
    Notice, these license terms, and the disclaimer/limitation of
    liability set forth as paragraph 6 below, in the documentation
    and/or other materials provided with the distribution.  For
    the purposes of binary distribution the "Copyright Notice"
    refers to the following language: "Copyright (c) 1999-2001
    Sendmail, Inc.  All rights reserved."

5.  Neither the name of Sendmail, Inc. nor its contributors may be
    used to endorse or promote products derived from this Software
    without specific prior written permission.  The name "sendmail"
    is a registered trademark of Sendmail, Inc.

6.  Disclaimer/Limitation of Liability: THIS SOFTWARE IS PROVIDED
    BY SENDMAIL, INC. AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
    PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL SENDMAIL, INC. OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
    STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
    ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
    ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.


#include <stdio.h>                      /* standard C I/O definitions */
#include <stdlib.h>                     /* standard C library definitions */
#include <string.h>                     /* standard C string definitions */
#include <assert.h>                     /* standard C assertion definitions */
#include <sys/stat.h>
#include <errno.h>
#include "db.h"				/* Berkely Database API */

#if	defined(STANDALONE_TEST)
#include "msdb-standalone.h"			

#include "msdb_odbc.h"			/* ODBC-specific stuff */
#include <sql.h>
#include <sqlext.h>

/* BEGIN variables protected by instantiate_lock mutex */
static pthread_mutex_t instantiate_lock = PTHREAD_MUTEX_INITIALIZER;
static int instantiated_p = 0;
static pthread_key_t per_thread_odbc_env_key;
char *ds_name = NULL, *ds_user = NULL, *ds_pass = NULL; /* Cannot be static */
/* END variables protected by instantiate_lock mutex */

static ODBC_ENV_t *msdb_instantiate_per_thread(void);
static void msdb_destroy_odbc(void *);
void dumpodbclog(SQLHENV, SQLHDBC, SQLHSTMT);

/*  */
 
SUCCFAIL				/* R: SUCC if database instantiated else FAIL */
msdb_instantiate (
/* PARAMETERS */
  char *db_path,			/* I: base path to databases */
  MS_DB_CONTEXT_t ** db,		/* O: new database context */
  MSDB_VERBOSITY_t verbosity,		/* I: verbosity level */
  char ** errmsg			/* O: error message if FAIL */
/* END PARAMETERS */
)
{
/* VARIABLES */
    char *c;				/* temp string pointer */
    SUCCFAIL return_value;		/* return value of function */
    SQLCHAR error[ODBC_MAXERRLEN+1];	/* ODBC error string */
    MS_DB_CONTEXT_t *local_db;		/* working copy of db context */
    ODBC_ENV_t *odbc_env;		/* local ODBC environment struct */
/* END VARIABLES */

    assert (db != NULL);
    pthread_mutex_lock(&instantiate_lock);
    /* XXXYYYXXX We only get one crack at this.... */
    if (instantiated_p) {
	abort();			/* XXXYYYXXX */
    } else {
	instantiated_p = 1;
    }
    pthread_key_create(&per_thread_odbc_env_key, msdb_destroy_odbc);
    /* XXXYYYZZZ name, user, & passwd should come from config file */
#if	defined(STANDALONE_TEST)
    /*
    ** ds_name, ds_user, and ds_pass have been set already by main().
    ** No mutex protection was necessary for them.
    */
#else	/* STANDALONE_TEST */
YOU LOSE
#endif	/* STANDALONE_TEST */
    pthread_mutex_unlock(&instantiate_lock);

/* ZZZ stuff missing */ 

/* : allocate and init DB_ENV structure */
    if ((odbc_env = (ODBC_ENV_t *) fs_get(sizeof (ODBC_ENV_t))) == NULL) {
	cleanup_return(FAIL);
    }
    memset(odbc_env, 0, sizeof(ODBC_ENV_t));

    local_db->db_env = (DB_ENV *) odbc_env;

/* : cleanup and return */
 CLEANUP:
    if (return_value == SUCC) {
	*db = local_db;
    }
    else {                              /* failure */
	msdb_destroy (&local_db);
	if (*errmsg == NULL) {          /* if null then assume internal err */
	    *errmsg = "Nasty error, whatever -SLF";
	}
    }
    return (return_value);
}
 
ODBC_ENV_t *
get_per_thread_odbc_env(void)
{
    ODBC_ENV_t	*odbc_env;

    if ((odbc_env = pthread_getspecific(per_thread_odbc_env_key)) == NULL) {
	odbc_env = msdb_instantiate_per_thread();
    }
    return odbc_env;
}


static ODBC_ENV_t *
msdb_instantiate_per_thread(void)
{
    SUCCFAIL status;
    ODBC_ENV_t *odbc_env;		/* Thread-local ODBC environment */
    SQLRETURN retcode;			/* ODBC return value */
    SQLCHAR sqlstate[6];		/* ODBC 5-char error string */
    SDWORD nativeerror;			/* ODBC native error code */
    SWORD errorlen;			/* ODBC error string length */
    SQLCHAR error[ODBC_MAXERRLEN+1];	/* ODBC error string */

    status = SUCC;

    odbc_env = fs_get(sizeof(ODBC_ENV_t));

    retcode = SQLAllocEnv(&(odbc_env->henv));
fprintf(stderr, "XXXYYYZZZ SQLAllocEnv retcode = %d\n", retcode);
dumpodbclog(odbc_env->henv, NULL, NULL);
    assert(SQL_SUCCEEDED(retcode));
fprintf(stderr, "XXXYYYZZZ allocated henv handle %d\n", (int) odbc_env->henv);
    /* XXX SQLSetEnvAttr(odbc_env->henv, SQL_ATTR_ODBC_VERSION,
      (SQLPOINTER) SQL_OV_ODBC2, 0 ); */
    retcode = SQLAllocConnect(odbc_env->henv,&odbc_env->hdbc);
    assert(SQL_SUCCEEDED(retcode));
    retcode = SQLSetConnectOption(odbc_env->hdbc, SQL_LOGIN_TIMEOUT, 15);
    assert(SQL_SUCCEEDED(retcode));
    /*
    ** XXXYYYZZZ What is the best option here? 
    ** According to Clustra, "SQL_TXN_REPEATABLE_READ: nothing else will do".
    */
    retcode = SQLConnect(odbc_env->hdbc, ds_name, SQL_NTS,
			 ds_user, SQL_NTS, ds_pass, SQL_NTS);
    if (retcode != SQL_SUCCESS && retcode != SQL_SUCCESS_WITH_INFO) {
	if (SQLError(odbc_env->henv, odbc_env->hdbc, 0,
		     sqlstate, &nativeerror, error, ODBC_MAXERRLEN,
		     &errorlen) != SQL_SUCCESS) {
	    /* No error in the connection level, so try at the env level */
	    SQLError(odbc_env->henv, 0, 0,
		     sqlstate, &nativeerror, error, ODBC_MAXERRLEN,
		     &errorlen);
	}
	fprintf(stderr, "XXXYYYZZZ SQLConnect retcode = %d\n", retcode);
	fprintf(stderr, "XXXYYYZZZ native error = %ld, error = %s\n", nativeerror, error);
	status = FAIL;
	goto cleanup;
    }
    /* odbc_env->autoc_p is set later */
    odbc_env->mytid = pthread_self();
    odbc_env->txn_p = 0;
    /* XXXYYYZZZ Note: we're not checking for out-of-memory here! */
    odbc_env->in_idbuf = fs_get(ODBC_MAXLEN_ID + 1);
    odbc_env->in_idbuflen = fs_get(sizeof(int));
    odbc_env->res_idbuf = fs_get(ODBC_MAXLEN_ID + 1);
    odbc_env->res_idbuflen = fs_get(sizeof(int));
    odbc_env->part = fs_get(sizeof(int));
    odbc_env->partmax = fs_get(sizeof(int));
    odbc_env->res_valbuf = fs_get(ODBC_MAXLEN_VAL + 1);
    odbc_env->res_valbuflen = fs_get(sizeof(int));

    /*
    ** Ah, yes, placement makes a _huge_ difference.  The SQL_AUTOCOMMIT
    ** option must be set after a connection has been successfully
    ** established.
    */
    retcode = SQLSetConnectOption(odbc_env->hdbc, SQL_AUTOCOMMIT,
				  SQL_AUTOCOMMIT_ON);
    odbc_env->autoc_p = 1;
    assert(retcode == SQL_SUCCESS);
    retcode = SQLSetConnectOption(odbc_env->hdbc, SQL_TXN_ISOLATION,
				  SQL_TXN_REPEATABLE_READ);
    assert(SQL_SUCCEEDED(retcode));

    /* save for future use */
    pthread_setspecific(per_thread_odbc_env_key, odbc_env);
fprintf(stderr, "XXXYYYZZZ thread %lx stashed odbc_env %lx\n", (unsigned long) pthread_self(), (unsigned long) odbc_env);
    status = SUCC;

 cleanup:
    if (status == FAIL) {
	/* XXX This error should be handled better, but how?  Syslog? */
	fprintf(stderr, "msdb_instantiate_per_thread() failed, prepare for assert problems!\n");
	return NULL;
    }
    return odbc_env;
}

void
msdb_destroy(MS_DB_CONTEXT_t **db)
{
    pthread_mutex_lock(&instantiate_lock);
    /* XXXYYYXXX We only get one crack at this.... */
    if (! instantiated_p) {
	abort();			/* XXXYYYXXX */
    } else {
	instantiated_p = 0;
    }
    pthread_mutex_unlock(&instantiate_lock);
    if (*db != NULL) {
/* : free up rest on context and context itself */
        fs_give ((void **) db);
    }
}

static void
msdb_destroy_odbc(void * p)
{
    ODBC_ENV_t *odbc_env = (ODBC_ENV_t *) p;

    assert(odbc_env != NULL);

    fs_give((void **) &odbc_env->in_idbuf);
    fs_give((void **) &odbc_env->in_idbuflen);
    fs_give((void **) &odbc_env->res_idbuf);
    fs_give((void **) &odbc_env->res_idbuflen);
    fs_give((void **) &odbc_env->part);
    fs_give((void **) &odbc_env->partmax);
    fs_give((void **) &odbc_env->res_valbuf);
    fs_give((void **) &odbc_env->res_valbuflen);
    /*
    ** XXX If statement handle caching is done, we'll need to free those
    ** data structures here.
    */
    SQLDisconnect(odbc_env->hdbc);
    SQLFreeConnect(odbc_env->hdbc);
    SQLFreeEnv(odbc_env->henv);
}

