        BAMS TECHNOLOGY LICENSE

The following license terms and conditions apply to this BAMS
software ("Software"), unless a different license is obtained from
Sendmail, Inc. at 6425 Christie Ave, Fourth Floor, Emeryville, CA 94608

=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

Copyright (c) 1999 - 2001
Sendmail, Inc.  All rights reserved.

License Terms:

Use, Modification and Redistribution (including distribution of
any modified or derived work) of the Software in source and binary
forms is permitted only if each of the following conditions is met:

1.  Redistributions of the Software and any modifications thereto
    or accompanying software that uses the Software are made at no
    charge beyond the reasonable cost of materials and delivery.

2.  Redistributions are accompanied by a complete copy of the Source
    Code or by an irrevocable offer to provide a complete copy of
    the Source Code for up to three years at the cost of materials
    and delivery.  Such redistributions must allow further use,
    modification, and redistribution of the Source Code under
    substantially the same terms as this license.  For the purposes
    of redistribution "Source Code" means the complete compilable
    and linkable source code of Software including all modifications
    and all accompanying software that uses the Software.  It does
    not include accompanying software modules or files that either
    (a) typically accompany the major components of the operating
    system on which the executable file runs, or (b) are distributed
    under another open source license, as defined by the Open Source
    Initiative (http://www.opensource.org).

3.  Redistributions of source code must retain the copyright notices
    as they appear in each source code file, these license terms,
    and the disclaimer/limitation of liability set forth as paragraph
    6 below.

4.  Redistributions in binary form must reproduce the Copyright
    Notice, these license terms, and the disclaimer/limitation of
    liability set forth as paragraph 6 below, in the documentation
    and/or other materials provided with the distribution.  For
    the purposes of binary distribution the "Copyright Notice"
    refers to the following language: "Copyright (c) 1999-2001
    Sendmail, Inc.  All rights reserved."

5.  Neither the name of Sendmail, Inc. nor its contributors may be
    used to endorse or promote products derived from this Software
    without specific prior written permission.  The name "sendmail"
    is a registered trademark of Sendmail, Inc.

6.  Disclaimer/Limitation of Liability: THIS SOFTWARE IS PROVIDED
    BY SENDMAIL, INC. AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
    PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL SENDMAIL, INC. OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
    STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
    ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
    ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.


/*
** XXXYYYZZZ Well, this is an interesting thing.  If I try using
** "DECLARE foo AS CURSOR FOR SELECT...", SQLExecute(), then
** "FETCH 100 FROM foo", the PostgreSQL driver complains:
**      S1000: [unixODBC]Error while executing the query;
**      Unexpected protocol character from backend (send_query)
**
** I'm not quite certain what the deal is, but I'm confused.
**
** I don't have my ODBC book handy (don't want to break it out of my
** suitcase here at the MSP gate), so I'm going to rely on the PostgreSQL
** driver's use of DECLARE/FETCH cursors to do what I need it to do,
** just so I can progress forward while I wait.
**
** NOTE: relying on this cursor feature is, in general, not a good thing.
*/

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include "db.h"                         /* Berkeley DB database API */

#include <sql.h>
#include <sqlext.h>

#include "msdb_low_level.h"
#include "msdb.h"

#include "msdb_odbc.h"

#include <syslog.h>	/* XXXYYYXXX debugging only */

#define	XXX_MAXERRLEN		500
#define XXX_MAXQUERYLEN		500
#define TRY_AGAIN               -666
#include <errno.h>
#if	defined(EREMOTEIO)
#define XXX_BUMMER_ERROR 	EREMOTEIO
#elif	defined(ECANCELED)
#define	XXX_BUMMER_ERROR	ECANCELED
#else
#define	XXX_BUMMER_ERROR	ERANGE
#endif

int msdb__transaction_end_2 (ODBC_ENV_t *, UWORD);
static int encode_binary(const unsigned char *, int, unsigned char *);
static int decode_binary(const unsigned char *, int, unsigned char **, unsigned char *);
static int
prep_select(char *tablename, ODBC_ENV_t *odbc_env, SQLHSTMT *hstmtp,
	    char *queryfmt,
	    char *in_id, int *in_idlen, char *out_id, int *out_idlen,
	    int *out_part, int *out_partmax, char *out_val, int *out_vallen,
	    int update);
static int
do_multipart_fetch(ODBC_ENV_t *odbc_env, SQLHSTMT hstmt, char **dstid,
		   int *dstidlen, char **dstval, int *dstvallen,
		   char *idbuf, int *idbuflen,
		   int *partp, int *partmaxp,
		   char *valbuf, int *valbuflen);

/* XXXYYYXXX prototypes for debugging functions */
void dumpodbclog(SQLHENV, SQLHDBC, SQLHSTMT); /* XXXYYYXXX debugging */ 

#if	defined(STANDALONE_TEST)
void *
fs_get(size_t n)
{
    return malloc(n);
}

void
fs_give(void **p)
{
    if ((p) != NULL) {
	free(*p);
	*p = NULL;
    }
}
#endif	/* STANDALONE_TEST */

/* END PRIVATE DEPENDENCIES */

int					/* R: 0 if ok, else errno */
msdb__record_put_flags (		/* D: put record in database */
/* PARAMETERS */
  DB *db,				/* I: database table but is really
				         *  : table name, (char *). */
  DB_TXN *tid,				/* I: transaction id but is really
					 *  : (ODBC_ENV_t *). */
  void *key,				/* I: key */
  unsigned int key_len,			/* I: length of key */
  void *val,				/* I: value */
  unsigned int val_len,			/* I: length of value */
  u_int32_t flags                       /* I: db put() flags */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;
    ODBC_ENV_t *odbc_env;
    char *tablename;
    SQLHSTMT hstmti, hstmtu, hstmt_lock; /* stmt handles for insert, update,
					    and select */
    SQLRETURN retcode;
    int num_rows_in_db=0;
    int retry_p;
    unsigned char *encval;
    int encvallen;
    int size;
    unsigned int part, partmax, bytes_remaining, bytes_towrite;
    char query[XXX_MAXQUERYLEN];
    odbc_error_s es;
/* END VARIABLES */

/* : sanity checks */
    assert (db != NULL);
    assert (tid != NULL);
    assert (key != NULL);
    assert ((val != NULL && val_len != 0) || (val == NULL && val_len == 0));
    assert (flags == 0 || flags == DB_NOOVERWRITE);
    if (key_len > ODBC_MAXLEN_ID) {
fprintf(stderr, "XXXYYYZZZ BUMMER record_put 1\n");
	return XXX_BUMMER_ERROR;
    }

/* : initialization */
    
    odbc_env = get_per_thread_odbc_env();
    assert(odbc_env != NULL);
    assert(odbc_env->txn_p);
    retry_p = 0;
    hstmtu = 0;
    hstmti = 0;
    hstmt_lock=0;
    tablename = (char *) db;
fprintf(stderr, "TRACE: msdb__record_put_flags: thread 0x%lx: table = %s ... key = 0x%lx for %d, val = 0x%lx for %d, flags = %d\n", (unsigned long) pthread_self(), tablename, (unsigned long) key, key_len, (unsigned long) val, val_len, flags);

 try_again:
    
    /*
    ** XXXYYYXXX !@#$!@#$! Not all of these things generate errors that
    ** dumpodbclog() can see.  For example, a syntax error in the SQL
    ** query will cause SQLPrepare() to puke, but it doesn't necessarily
    ** have an error that dupodbclog() can see.  !@#$!@#$!
    */

    /* Before going any further, encode the key and val */
    if ((*odbc_env->in_idbuflen =
	 encode_binary((const unsigned char *) key, key_len,
		       odbc_env->in_idbuf)) > ODBC_MAXLEN_ID) {
	r = EINVAL;
	goto cleanup;
    }
    if ((encval = (unsigned char *) fs_get(val_len * 2)) == NULL) {
	r = errno;
	goto cleanup;
    }
    encvallen = encode_binary((const unsigned char *) val, val_len, encval);

    /* calculate parts and offsets */
    /* do this here so we might decide on an inteligent update strategy
       later */
    part = 1;
    partmax = encvallen / ODBC_MAXLEN_VAL;
    if (partmax * ODBC_MAXLEN_VAL == encvallen)	/* Divided evenly? */
      partmax = encvallen / ODBC_MAXLEN_VAL;
    else
      partmax++;			/* Incr for partial block */
    if (partmax == 0)			/* Key was zero length? */
      partmax = 1;
    bytes_remaining = encvallen;

    if (bytes_remaining > ODBC_MAXLEN_VAL)	
      bytes_towrite = ODBC_MAXLEN_VAL;
    else
      bytes_towrite = bytes_remaining;

    /*
    ** First, figure out if there's already a row(s) there. And lock it
    ** if it exists.
    */
    
    retcode = SQLAllocStmt(odbc_env->hdbc, &hstmt_lock);
    if (retcode != SQL_SUCCESS) {
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt_lock);
        fprintf(stderr, "XXXYYYZZZ BUMMER record_put 60\n");
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    
    snprintf(query, XXX_MAXQUERYLEN,
	     "UPDATE %s SET partmax=partmax WHERE id = ? and part=1", tablename);
    retcode = SQLPrepare(hstmt_lock, query, SQL_NTS);
    if (retcode != SQL_SUCCESS) {
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt_lock);
        fprintf(stderr, "XXXYYYZZZ BUMMER record_put 61 ret=%d\n",retcode);
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    
    /* Bind input parameters */
    retcode = SQLBindParameter(hstmt_lock, 1, SQL_PARAM_INPUT, SQL_C_CHAR,
			       SQL_CHAR, *odbc_env->in_idbuflen, 0,
			       (SQLPOINTER) odbc_env->in_idbuf,
			       *odbc_env->in_idbuflen,
			       (SQLINTEGER *) odbc_env->in_idbuflen);
    if (retcode != SQL_SUCCESS) {
	fprintf(stderr, "XXXYYYXXX SQLBindParameter #1 failed\n");
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt_lock);
        fprintf(stderr, "XXXYYYZZZ BUMMER record_put 62\n");
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    
    /* execute update on first row of record to lock and find out if there
       allready are rows with this id in the db*/
    retcode = SQLExecute(hstmt_lock);
    if (retcode != SQL_SUCCESS) {
	if (get_native_error(hstmt_lock,&es)<0) {
	    dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt_lock);
            fprintf(stderr, "XXXYYYZZZ BUMMER record_put 63\n");
	    r = XXX_BUMMER_ERROR;
	} else {
	    if (es.nativeerror == CLUSTRA_TIMEOUT ||
		es.nativeerror == CLUSTRA_UPGRADE_LOCK) {
	       r=ODBC_TIMEOUT;
	    } else {
                print_error(es);
                fprintf(stderr, "XXXYYYZZZ BUMMER record_put 63-2\n");
                r = XXX_BUMMER_ERROR;
	    }
	}
	goto cleanup;
    }

    retcode = SQLRowCount(hstmt_lock,(long*)&num_rows_in_db);  
    if (retcode != SQL_SUCCESS) {
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt_lock);
        fprintf(stderr, "XXXYYYZZZ BUMMER record_put 64\n");
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }

    if (num_rows_in_db > 0) {
        /* if we don't want to do an overwrite, return */
	if (flags == DB_NOOVERWRITE) {
	    r = DB_KEYEXIST;
	    goto cleanup;
	}
        
        if ((r = msdb__record_del_parts(db, tid, key, key_len,2)) != 0) {
            /*r= XXX_BUMMER_ERROR;*/
            goto cleanup;
	}
    }

    if (num_rows_in_db > 0) {
	/* do an update on the first row to maintain lock on this row 
	 * and update to new value
	 */
	retcode = SQLAllocStmt(odbc_env->hdbc, &hstmtu);
	if (retcode != SQL_SUCCESS) {
	    dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmtu);
	    fprintf(stderr, "XXXYYYZZZ BUMMER record_put 2\n");
	    r = XXX_BUMMER_ERROR;
	    goto cleanup;
	}
	
	snprintf(query, XXX_MAXQUERYLEN,
		 "UPDATE %s SET partmax=?,val=? WHERE id = ? and part=1", tablename);
	retcode = SQLPrepare(hstmtu, query, SQL_NTS);
	if (retcode != SQL_SUCCESS) {
	    dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmtu);
	    fprintf(stderr, "XXXYYYZZZ BUMMER record_put 3\n");
	    r = XXX_BUMMER_ERROR;
	    goto cleanup;
	}
	
	/* Bind input parameters */
	size = sizeof(partmax);
	retcode = SQLBindParameter(hstmtu, 1, SQL_PARAM_INPUT, SQL_C_SLONG,
				   SQL_INTEGER, SQL_NUMERIC, 0, (SQLPOINTER) &partmax,
				   size, (SQLINTEGER *) &size);
	if (retcode != SQL_SUCCESS) {
	    fprintf(stderr, "XXXYYYXXX SQLBindParameter #1 failed\n");
	    dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmtu);
	    fprintf(stderr, "XXXYYYZZZ BUMMER record_put 12-2\n");
	    r = XXX_BUMMER_ERROR;
	    goto cleanup;
	}
	
	retcode = SQLBindParameter(hstmtu, 2, SQL_PARAM_INPUT, SQL_C_CHAR,
				   SQL_CHAR, bytes_towrite, 0,
				   (SQLPOINTER) (encval + ((part - 1) * ODBC_MAXLEN_VAL)),
				   bytes_towrite,
				   (SQLINTEGER *) &bytes_towrite);
	if (retcode != SQL_SUCCESS) {
	    fprintf(stderr, "XXXYYYXXX SQLBindParameter #2 failed\n");
	    dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmtu);
	    fprintf(stderr, "XXXYYYZZZ BUMMER record_put 15-2\n");
	    r = XXX_BUMMER_ERROR;
	    goto cleanup;
	}
	
	retcode = SQLBindParameter(hstmtu, 3, SQL_PARAM_INPUT, SQL_C_CHAR,
				   SQL_CHAR, *odbc_env->in_idbuflen, 0,
				   (SQLPOINTER) odbc_env->in_idbuf,
				   *odbc_env->in_idbuflen,
				   (SQLINTEGER *) odbc_env->in_idbuflen);
	if (retcode != SQL_SUCCESS) {
	    fprintf(stderr, "XXXYYYXXX SQLBindParameter #3 failed\n");
	    dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmtu);
	    fprintf(stderr, "XXXYYYZZZ BUMMER record_put 4-2\n");
	    r = XXX_BUMMER_ERROR;
	    goto cleanup;
	}
	
	/* execute update on first row of record */
	retcode = SQLExecute(hstmtu);
	if (retcode != SQL_SUCCESS) {
	    if (get_native_error(hstmtu,&es) < 0) {
		dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmtu);
		fprintf(stderr, "XXXYYYZZZ BUMMER record_put 6\n");
		r = XXX_BUMMER_ERROR;
	    } else {
		if (es.nativeerror == CLUSTRA_TIMEOUT ||
		    es.nativeerror == CLUSTRA_UPGRADE_LOCK) {
		    r=ODBC_TIMEOUT;
		} else {
		    print_error(es);
		    fprintf(stderr, "XXXYYYZZZ BUMMER record_put 6-2\n");
		    r = XXX_BUMMER_ERROR;
		}
	    }
	    goto cleanup;
	}
	part++;
	bytes_remaining -= bytes_towrite;
    }

    /*
    ** if there is only one row in new record  and this is an update,
    ** stop here
    */
    if (partmax < 2 && num_rows_in_db > 0) {
	r=0;
	goto cleanup;
    }
    
    /*
    ** OK, now it should be safe to insert rows (if any)
    ** (if update, rows 2 - partmax).
    */
    
    retcode = SQLAllocStmt(odbc_env->hdbc, &hstmti);
    if (retcode != SQL_SUCCESS) {
        dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmti);
        fprintf(stderr, "XXXYYYZZZ BUMMER record_put 8\n");
        r = XXX_BUMMER_ERROR;
        goto cleanup;
    }
    
    snprintf(query, XXX_MAXQUERYLEN, "INSERT INTO %s values (?, ?, ?, ?)",
	     tablename);
    retcode = SQLPrepare(hstmti, query, SQL_NTS);
    if (retcode != SQL_SUCCESS) {
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmti);
        fprintf(stderr, "XXXYYYZZZ BUMMER record_put 9\n");
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    /* Bind input parameters */
    retcode = SQLBindParameter(hstmti, 1, SQL_PARAM_INPUT, SQL_C_CHAR,
			       SQL_CHAR, *odbc_env->in_idbuflen, 0,
			       (SQLPOINTER) odbc_env->in_idbuf,
			       *odbc_env->in_idbuflen,
			       (SQLINTEGER *) odbc_env->in_idbuflen);
    if (retcode != SQL_SUCCESS) {
	fprintf(stderr, "XXXYYYXXX SQLBindParameter #1 failed\n");
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmti);
        fprintf(stderr, "XXXYYYZZZ BUMMER record_put 10\n");
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    size = sizeof(part);
    retcode = SQLBindParameter(hstmti, 2, SQL_PARAM_INPUT, SQL_C_SLONG,
			       SQL_INTEGER, SQL_NUMERIC, 0, (SQLPOINTER) &part,
			       size, (SQLINTEGER *) &size);
    if (retcode != SQL_SUCCESS) {
	fprintf(stderr, "XXXYYYXXX SQLBindParameter #2 failed\n");
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmti);
        fprintf(stderr, "XXXYYYZZZ BUMMER record_put 11\n");
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    size = sizeof(partmax);
    retcode = SQLBindParameter(hstmti, 3, SQL_PARAM_INPUT, SQL_C_SLONG,
			       SQL_INTEGER, SQL_NUMERIC, 0, (SQLPOINTER) &partmax,
			       size, (SQLINTEGER *) &size);
    if (retcode != SQL_SUCCESS) {
	fprintf(stderr, "XXXYYYXXX SQLBindParameter #3 failed\n");
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmti);
        fprintf(stderr, "XXXYYYZZZ BUMMER record_put 12\n");
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    
    /*
    ** DEV ATTN Because the starting offset of params #4 varies, it
    ** needs to be bound in each iteration of the loop below.
    */
 
    while (bytes_remaining > 0) {        
	if (bytes_remaining > ODBC_MAXLEN_VAL)	
          bytes_towrite = ODBC_MAXLEN_VAL;
	else
          bytes_towrite = bytes_remaining;
	assert(bytes_towrite > 0);
        
        retcode = SQLBindParameter(hstmti, 4, SQL_PARAM_INPUT, SQL_C_CHAR,
                                   SQL_CHAR, bytes_towrite, 0,
                                   (SQLPOINTER) (encval + ((part - 1) * ODBC_MAXLEN_VAL)),
                                   bytes_towrite,
                                   (SQLINTEGER *) &bytes_towrite);
        if (retcode != SQL_SUCCESS) {
            fprintf(stderr, "XXXYYYXXX SQLBindParameter #4 failed\n");
            dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmti);
            fprintf(stderr, "XXXYYYZZZ BUMMER record_put 15\n");
            r = XXX_BUMMER_ERROR;
            goto cleanup;
	}

        /* XXXYYYXX Check for error returns for these bindings? */
        retcode = SQLExecute(hstmti);
        if (retcode == SQL_SUCCESS) {
            r = 0;
        } else if (retcode == SQL_SUCCESS_WITH_INFO) {
            dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmti);
            r = 0;
	} else {
	    if (get_native_error(hstmti,&es) < 0) {
	       dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmti);
               fprintf(stderr, "XXXYYYZZZ BUMMER record_put 16\n");
               r = XXX_BUMMER_ERROR;
	    } else {
	       if (es.nativeerror == CLUSTRA_TIMEOUT ||
		   es.nativeerror == CLUSTRA_UPGRADE_LOCK) {
		   r = ODBC_TIMEOUT;
	       /* we can risk there beeing another insert that came
               right before this one, in that case retry. */
	       } else if (es.nativeerror == CLUSTRA_KEY_EXISTS &&
			  part == 1 && num_rows_in_db < 1) {
		  r = TRY_AGAIN;
	       } else {
                   print_error(es);
                   r = XXX_BUMMER_ERROR;
	       }
            }
            goto cleanup;
	}
        part++;
        bytes_remaining -= bytes_towrite;
    }

    r = 0;

/* : cleanup and return success */
 cleanup:
    if (hstmt_lock != 0) SQLFreeStmt(hstmt_lock, SQL_DROP);
    if (hstmtu != 0) SQLFreeStmt(hstmtu, SQL_DROP);
    if (hstmti != 0) SQLFreeStmt(hstmti, SQL_DROP);
    if (encval != NULL) fs_give((void **) &encval);
    if (r == TRY_AGAIN)
	goto try_again;
    return (r);
}

/* END FUNCTION: msdb__record_put_flags */

/*  */

/* FUNCTION: msdb__record_get */

/* SYNOPSIS
 * Get a value identified by key from the database.
 * Outputed value is nul terminated.
 * END SYNOPSIS */

/* NOTES
 * If database stores duplicates, then only the first is returned.
 * Memory for value is allocated.  Caller is responsible for freeing it.
 * END NOTES */

int                                     /* R: 0 if ok, else errno */
msdb__record_get (                      /* D: get record from database */
  DB *db,				/* I: database table */
  DB_TXN *tid,				/* I: transaction id */
  void *key,				/* I: key */
  unsigned int key_len,			/* I: length of key data */
  void **val,				/* O: value */
  unsigned int *val_len			/* O: length of value data */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;                              /* return value */
    ODBC_ENV_t *odbc_env;
    char *tablename;
    SQLHSTMT hstmt;
    SQLRETURN retcode;
    char *resid;
    int residlen;
    char *queryfmt;
    char res_idbuf[ODBC_MAXLEN_ID+1], res_valbuf[ODBC_MAXLEN_VAL+1];
    int res_idbuflen, res_valbuflen;
    unsigned int part, partmax;
    odbc_error_s es;
/* END VARIABLES */

/* : sanity checks */
    assert (db != NULL);
    assert (key != NULL);
    assert (key_len != 0);
    assert (val != NULL);
    assert (val_len != NULL);
    if (key_len > ODBC_MAXLEN_ID) {
	return XXX_BUMMER_ERROR;
    }
    odbc_env = get_per_thread_odbc_env();
    assert(odbc_env != NULL);
    /*
    ** XXXYYYZZZ I'm going to relax the restriction that a get operation
    ** XXXYYYZZZ must be executed within a transaction context.  I'm not
    ** XXXYYYZZZ quite certain what effect that'll have with non-autocommit
    ** XXXYYYZZZ ODBC connections, but I'll soon find out.
    ** 
    ** assert(odbc_env->txn_p);
    */
    if ((*odbc_env->in_idbuflen =
	 encode_binary((const unsigned char *) key, key_len,
		       odbc_env->in_idbuf)) > ODBC_MAXLEN_ID) {
	r = EINVAL;
	goto cleanup;
    }
    hstmt = 0;
    res_idbuflen = 0;
    part = 0;
    partmax = 0;
    res_valbuflen = 0;

    tablename = (char *) db;
fprintf(stderr, "TRACE: msdb__record_get: thread 0x%lx: table = %s ... key = 0x%lx for %d\n", (unsigned long) pthread_self(), tablename, (unsigned long) key, key_len);
    queryfmt = "SELECT * FROM %s WHERE id = ? ORDER BY id, part";
    r = prep_select(tablename, odbc_env, &hstmt, queryfmt,
		    odbc_env->in_idbuf, odbc_env->in_idbuflen,
		    res_idbuf, &res_idbuflen, &part, &partmax,
		    res_valbuf, &res_valbuflen,0);
    if (r != 0) {
	goto cleanup;
    }
		     
    retcode = SQLExecute(hstmt);
    if (retcode != SQL_SUCCESS) {
        if(get_native_error(hstmt, &es) < 0) {
            dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
            fprintf(stderr,"SQLExecute msdb__record_get 90\n");
            r = XXX_BUMMER_ERROR;
	} else {
            if (es.nativeerror == CLUSTRA_TIMEOUT ||
		es.nativeerror == CLUSTRA_UPGRADE_LOCK) {
		r = ODBC_TIMEOUT;
            } else {
                print_error(es);
                r = XXX_BUMMER_ERROR;
                fprintf(stderr,"SQLExecute msdb__record_get 90-2\n"); 
	    }
	}	
        goto cleanup;
    }

    r = do_multipart_fetch(odbc_env, hstmt, NULL, 0, &resid, &residlen,
			   res_idbuf, &res_idbuflen, &part, &partmax,
			   res_valbuf, &res_valbuflen);
    if (r == 0) {
/* : DOC - we tack on a null char at end of string for convenience of caller */
	*(resid + residlen) = '\0';
	*val = resid;
	*val_len = residlen;
    } else {
        if (r != ODBC_TIMEOUT)
	    r = DB_NOTFOUND;
    }
    
 cleanup:
    if (hstmt != 0)
      SQLFreeStmt(hstmt, SQL_DROP);
    return (r);
}

/* END FUNCTION: msdb__record_get */

/*  */

/* FUNCTION: msdb__record_del */

/* SYNOPSIS
 * END SYNOPSIS */

/* NOTES
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__record_del (			/* D: delete record from database */
/* PARAMETERS */
  DB *db,				/* I: database table */
  DB_TXN *tid,				/* I: transaction id */
  void *key,				/* I: key */
  unsigned int key_len			/* I: length of key data */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;                              /* generic return value */
    ODBC_ENV_t *odbc_env;
    char *tablename;
    SQLHSTMT hstmt;
    SQLRETURN retcode;
    char query[XXX_MAXQUERYLEN];
    odbc_error_s es;
/* END VARIABLES */

/* : sanity checks */
    assert (db != NULL);
    assert (key != NULL);
    assert (key_len != 0);
    if (key_len > ODBC_MAXLEN_ID) {
	return XXX_BUMMER_ERROR;
    }

/* : initialization */
    hstmt = 0;
    odbc_env = get_per_thread_odbc_env();
    assert(odbc_env != NULL);
    assert(odbc_env->txn_p);
    if ((*odbc_env->in_idbuflen =
	 encode_binary((const unsigned char *) key, key_len,
		       odbc_env->in_idbuf)) > ODBC_MAXLEN_ID) {
	r = EINVAL;
	goto cleanup;
    }
 
    retcode = SQLAllocStmt(odbc_env->hdbc, &hstmt);
    if (retcode != SQL_SUCCESS) {
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    tablename = (char *) db;
fprintf(stderr, "TRACE: msdb__record_del: thread 0x%lx: table = %s ... key = 0x%lx for %d\n", (unsigned long) pthread_self(), tablename, (unsigned long) key, key_len);
    snprintf(query, XXX_MAXQUERYLEN,
	     "DELETE FROM %s where id = ?",
	     tablename);
    retcode = SQLPrepare(hstmt, query, SQL_NTS);
    if (retcode != SQL_SUCCESS) {
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }

    /* XXXYYYXX Check for error returns for these bindings? */
    retcode = SQLBindParameter(hstmt, 1, SQL_PARAM_INPUT, SQL_C_CHAR,
			       SQL_CHAR, *odbc_env->in_idbuflen, 0,
			       (SQLPOINTER) odbc_env->in_idbuf,
			       *odbc_env->in_idbuflen,
			       (SQLINTEGER *) odbc_env->in_idbuflen);
    if (retcode != SQL_SUCCESS) {
	fprintf(stderr, "XXXYYYXXX SQLBindParameter update #1 failed\n");
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    retcode = SQLExecute(hstmt);
    if (retcode == SQL_SUCCESS) {
	/* :- return success if key deleted or didn't exist */
	r = 0;
    } else if (retcode == SQL_SUCCESS_WITH_INFO) {
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
	/* :- return success if key deleted or didn't exist */
	r = 0;
    } else {
        if (get_native_error(hstmt, &es) < 0) {
            dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
            fprintf(stderr,"SQLExecute msdb_record_del\n");
            r = XXX_BUMMER_ERROR;
	} else {
            if (es.nativeerror == CLUSTRA_TIMEOUT ||
		es.nativeerror == CLUSTRA_UPGRADE_LOCK) {
		r = ODBC_TIMEOUT;
            } else {
                print_error(es);
                fprintf(stderr,"SQLExecute msdb_record_del-2\n");
                r = XXX_BUMMER_ERROR;
	    }
	}
    }
 
/* : cleanup and return err */
 cleanup:
    if (hstmt != 0) SQLFreeStmt(hstmt, SQL_DROP);
    return (r);
}

/* END FUNCTION: msdb__record_del */

int					/* R: 0 if ok, else errno */
msdb__record_del_parts (		/* D: delete record from database where part>=part_min*/
/* PARAMETERS */
  DB *db,				/* I: database table */
  DB_TXN *tid,				/* I: transaction id */
  void *key,				/* I: key */
  unsigned int key_len,			/* I: length of key data */
  unsigned int part_min                 /* I: delete all parts that is bigger or equal than part_min */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;                              /* generic return value */
    ODBC_ENV_t *odbc_env;
    char *tablename;
    SQLHSTMT hstmt;
    SQLRETURN retcode;
    char query[XXX_MAXQUERYLEN];
    odbc_error_s es;
/* END VARIABLES */

/* : sanity checks */
    assert (db != NULL);
    assert (key != NULL);
    assert (key_len != 0);
    if (key_len > ODBC_MAXLEN_ID) {
	return XXX_BUMMER_ERROR;
    }

/* : initialization */
    hstmt = 0;
    odbc_env = get_per_thread_odbc_env();
    assert(odbc_env != NULL);
    assert(odbc_env->txn_p);
    if ((*odbc_env->in_idbuflen =
	 encode_binary((const unsigned char *) key, key_len,
		       odbc_env->in_idbuf)) > ODBC_MAXLEN_ID) {
	r = EINVAL;
	goto cleanup;
    }
 
    retcode = SQLAllocStmt(odbc_env->hdbc, &hstmt);
    if (retcode != SQL_SUCCESS) {
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    tablename = (char *) db;
    snprintf(query, XXX_MAXQUERYLEN,
	     "DELETE FROM %s where id = ? and part>=%d",
	     tablename,part_min);
    retcode = SQLPrepare(hstmt, query, SQL_NTS);
    if (retcode != SQL_SUCCESS) {
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }

    /* XXXYYYXX Check for error returns for these bindings? */
    retcode = SQLBindParameter(hstmt, 1, SQL_PARAM_INPUT, SQL_C_CHAR,
			       SQL_CHAR, *odbc_env->in_idbuflen, 0,
			       (SQLPOINTER) odbc_env->in_idbuf,
			       *odbc_env->in_idbuflen,
			       (SQLINTEGER *) odbc_env->in_idbuflen);
    if (retcode != SQL_SUCCESS) {
	fprintf(stderr, "XXXYYYXXX SQLBindParameter update #1 failed\n");
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    retcode = SQLExecute(hstmt);
    if (retcode == SQL_SUCCESS) {
        /* :- return success if key deleted or didn't exist */
	r = 0;
    } else if (retcode == SQL_SUCCESS_WITH_INFO) {
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
	/* :- return success if key deleted or didn't exist */
	r = 0;
    } else {
        if (get_native_error(hstmt, &es) < 0) {
            dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
            r = XXX_BUMMER_ERROR;
	} else {
            if (es.nativeerror == CLUSTRA_TIMEOUT ||
		es.nativeerror == CLUSTRA_UPGRADE_LOCK) {
		r = ODBC_TIMEOUT;
            } else {
                print_error(es);
                r = XXX_BUMMER_ERROR;
	    }
	}
    }
 
/* : cleanup and return err */
 cleanup:
    if (hstmt != 0) SQLFreeStmt(hstmt, SQL_DROP);
    return (r);
}

/* END FUNCTION: msdb__record_del_parts */

/*  */

/* FUNCTION: msdb__record_exists */

/* SYNOPSIS
 * Check for the existance of a key (and optionally the value) in a database.
 * END SYNOPSIS */

/* NOTES
 * END NOTES */

int					/* R: 0 if ok, DB_NOTFOUND if not exists, else errno */
msdb__record_exists (
/* PARAMETERS */
  DB * db,				/* I: message store database context */
  DB_TXN *tid,				/* I: transaction id */
  void * key,				/* I: key to check for */
  unsigned int keylen,			/* I: length of key */
  void * val,				/* I: value to check for (if non-null) */
  unsigned int vallen			/* I: length of value (or 0) */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;				/* return value */
    char *dbval;
    int dbvallen;
/* END VARIABLES */

/* : sanity checks */
    assert (db != NULL);
    assert (key != NULL);

/* : initialization */
    dbval = NULL;
    dbvallen = 0;

fprintf(stderr, "TRACE: msdb__record_exists: thread 0x%lx: table = %s ... key = 0x%lx for %d, val = 0x%lx for %d\n", (unsigned long) pthread_self(), (char *) db, (unsigned long) key, keylen, (unsigned long) val, vallen);
/* : retrieve value from database */
    r = msdb__record_get(db, tid, key, keylen, (void **) &dbval, &dbvallen);
    if (r != 0) {
	goto cleanup;
    }

/* : ensure value matches given value */
    if (val != NULL) {
	if (dbvallen != vallen) {
	    r = 1;
	    goto cleanup;
	}
/* : CLAIM - length of val and dbval.size are the same and the strings can be compared */
	if (memcmp (val, dbval, dbvallen) != 0) {
	    r = 1;
	    goto cleanup;
	}
    }

/* : cleanup and return success */
 cleanup:
    if (dbval != NULL) fs_give((void **) &dbval);
    return (r);
}

/* END FUNCTION: msdb__record_exists */
 
/*  */

/* FUNCTION: msdb__record_lock */

/* SYNOPSIS
 * END SYNOPSIS */

/* NOTES
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__record_lock (			/* D: lock record in database */
/* PARAMETERS */
/* END PARAMETERS */
)
{
/* VARIABLES */
/* END VARIABLES */

/* : sanity checks */

/* : initialization */

/* : cleanup and return */
    return (0);
}

/* END FUNCTION: msdb__record_lock */


/*  */

/* FUNCTION: msdb__record_unlock */

/* SYNOPSIS
 * END SYNOPSIS */

/* NOTES
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__record_unlock (			/* D: unlock record in databse */
/* PARAMETERS */
/* END PARAMETERS */
)
{
/* VARIABLES */
/* END VARIABLES */

/* : sanity checks */

/* : initialization */

/* : cleanup and return */
    return (0);
}

/* END FUNCTION: msdb__record_unlock */


/*  */

/* FUNCTION: msdb__cursor_open */

/*
** XXXYYYXXX I have several worries about the cursor emulation that will
** be required to pull this off, but (for now) I'm going to act as if I
** don't give a damn and just write full steam ahead.  I don't know 
** enough about ODBC to try to solve these problems ahead of time, and
** I don't know how much behavior is database driver-specific, so that
** a solution for one (like my PostgreSQL experiment) won't work for
** another (like the upcoming Clustra evaluation).
**
** The main problem is dealing with the size of the data set returned
** by cursors and ODBC's implicit cursor handling.  What happens if I
** want to scan a large table via a cursor, and I use "SELECT * FROM
** table" as my cursor criteria, and the driver tries to slurp in the
** entire table to support its implicit cursor?  Or, almost as bad, it
** does async double buffering to slurp in the table faster than it is
** read?  What happens in the msdb__get_closest() if it does a "SELECT
** * FROM table WHERE id LIKE 'foo%'" and you get a few thousand 
** rows?
**
** From what I see, the ODBC 2.x answer is to use the SQL_MAX_ROWS
** connection option to limit the amount of data the driver slurps.
** But if you set that limit to N, and you're trying to read a cursor
** on a result set of more than N rows, then you need to remember what
** the last returned key was so that, after reading N rows, you can
** close the cursor and open a new one that starts at that last-read
** key.  Further complicating matters is a value where its partmax > 1,
** and the N row limit comes while in the middle of reading the
** constituent chunks of a value.  ICKY!
**
** So, for now, I'm going to assume that everything will behave with
** aplomb and that I don't have to worry.....
*/

/* SYNOPSIS
 * Create a cursor.
 * END SYNOPSIS */

/* NOTES
 * Records are implicitly locked by the transaction subsytem when tid != NULL
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__cursor_open (			/* D: create new cursor */
/* PARAMETERS */
  DB *db,				/* I: database table */
  DB_TXN *tid,				/* I: transaction id */
  DBC **cursor				/* O: cursor pointer */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;				/* generic return code */
    ODBC_ENV_t *odbc_env;
    ODBC_CURSOR_t *curstate;
/* END VARIABLES */

/* : sanity checks */
    assert (db != NULL);
    assert (cursor != NULL);

/* : initialization */
    odbc_env = get_per_thread_odbc_env();
    assert(odbc_env != NULL);
    if ((curstate = fs_get(sizeof(ODBC_CURSOR_t))) == NULL) {
	return 1;
    }

    /*
    ** XXXYYYZZZ I'm going to relax the restriction that a get operation
    ** XXXYYYZZZ must be executed within a transaction context.  I'm not
    ** XXXYYYZZZ quite certain what effect that'll have with non-autocommit
    ** XXXYYYZZZ ODBC connections, but I'll soon find out.
    ** ZZZ The only problem here is that we don't know at cursor opening
    ** ZZZ time if the cursor will be used for writing.  So....
    ** ZZZ Let the caller beware!
    **
    ** assert(odbc_env->txn_p);
    */

    curstate->odbc_env = odbc_env;
    curstate->ctable = (char *) db;	/* XXXYYYZZZ Is not dup'ing this string gonna cause trouble? */
    curstate->cstmt = 0;
    /* curstate->ccurname is currently unused */
    curstate->in_idbuf = fs_get(ODBC_MAXLEN_ID + 1);
    curstate->in_idbuflen = fs_get(sizeof(int));
    curstate->res_idbuf = fs_get(ODBC_MAXLEN_ID + 1);
    curstate->res_idbuflen = fs_get(sizeof(int));
    curstate->part = fs_get(sizeof(int));
    curstate->partmax = fs_get(sizeof(int));
    curstate->res_valbuf = fs_get(ODBC_MAXLEN_VAL + 1);
    curstate->res_valbuflen = fs_get(sizeof(int));
    *cursor = (DBC *) curstate;
    r = 0;

fprintf(stderr, "TRACE: msdb__cursor_open: thread 0x%lx: cursor = 0x%lx\n", (unsigned long) pthread_self(), (unsigned long) *cursor);
/* : cleanup and return */
    return (r);
}

/* END FUNCTION: msdb__cursor_open */


/*  */

/* FUNCTION: msdb__cursor_close */

/* SYNOPSIS
 * Close a cursor.
 * END SYNOPSIS */

/* NOTES
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__cursor_close (			/* D: close cursor */
/* PARAMETERS */
  DBC **cursor				/* U: cursor pointer */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;				/* generic return code */
    ODBC_CURSOR_t *curstate;
/* END VARIABLES */

/* : sanity checks */
    assert (cursor != NULL);

/* : initialization */
    r = 0;
    assert(cursor != NULL);
    curstate = (ODBC_CURSOR_t *) (*cursor);
    assert(curstate != NULL);
fprintf(stderr, "TRACE: msdb__cursor_close: thread 0x%lx: cursor = 0x%lx\n", (unsigned long) pthread_self(), (unsigned long) *cursor);

    if (curstate->cstmt != 0)
	SQLFreeStmt(curstate->cstmt, SQL_DROP);
    fs_give((void **) &curstate->in_idbuf);
    fs_give((void **) &curstate->in_idbuflen);
    fs_give((void **) &curstate->res_idbuf);
    fs_give((void **) &curstate->res_idbuflen);
    fs_give((void **) &curstate->part);
    fs_give((void **) &curstate->partmax);
    fs_give((void **) &curstate->res_valbuf);
    fs_give((void **) &curstate->res_valbuflen);
    *cursor = NULL;

/* : cleanup and return */
    return (r);
}

/* END FUNCTION: msdb__cursor_close */


/*  */

/* FUNCTION: msdb__cursor_get_exact */

/* SYNOPSIS
 * Outputed value is nul terminated.
 * END SYNOPSIS */

/* NOTES
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__cursor_get_exact (		/* D: set cursor to given key */
/* PARAMETERS */
  DBC *cursor,				/* I: cursor pointer */
  void *key,				/* I: key */
  unsigned int key_len,			/* I: length of key data */
  void **val,				/* O: value */
#ifdef	XXX_SLF_API_CHANGE_HACK
  unsigned int *val_len,		/* O: length of val data */
  int update                            /* I: create update cursor? */
#else	/* XXX_SLF_API_CHANGE_HACK */
  unsigned int *val_len			/* O: length of val data */
#endif	/* XXX_SLF_API_CHANGE_HACK */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;				/* generic return code */
    ODBC_CURSOR_t *curstate;
    char *tablename;
    SQLHSTMT hstmt;
    SQLRETURN retcode;
    char *resid;
    int residlen;
    char queryfmt[ODBC_MAXQUERYLEN];
    odbc_error_s es;
/* END VARIABLES */

/* : sanity checks */
    assert (cursor != NULL);
    assert (key != NULL);
    assert (key_len != 0);
    assert (val != NULL);
    assert (val_len != NULL);
    curstate = (ODBC_CURSOR_t *) cursor;
    /* XXXYYYZZZ not required for reading, eh? assert(odbc_env->txn_p); */
    if (curstate->ctable == NULL) {
	fprintf(stderr, "XXXYYYXXX msdb__cursor_get_exact called without prior open\n");
	syslog(LOG_CRIT, "XXXYYYXXX msdb__cursor_get_exact called without prior open");
	return 1;
    }
    if (curstate->cstmt != 0) {
	fprintf(stderr, "XXXYYYXXX msdb__cursor_get_exact called with previous active cursor\n");
	syslog(LOG_CRIT, "XXXYYYXXX msdb__cursor_get_exact called with previous active cursor");
	return 1;
    }

/* : initialization */
    if ((*curstate->in_idbuflen =
	 encode_binary((const unsigned char *) key, key_len,
				   curstate->in_idbuf)) > ODBC_MAXLEN_ID) {
	r = EINVAL;
	goto cleanup;
    }
    hstmt = 0;
    resid = NULL;
    residlen = 0;

    tablename = (char *) curstate->ctable;
fprintf(stderr, "TRACE: msdb__cursor_get_exact: thread 0x%lx: table = %s ... key = 0x%lx for %d, cursor = 0x%lx\n", (unsigned long) pthread_self(), tablename, (unsigned long) key, key_len, (unsigned long) cursor);
    snprintf(curstate->ccurname, ODBC_MAXLEN_CNAME, "mycursor0");
    snprintf(queryfmt, ODBC_MAXQUERYLEN,
	     "SELECT * FROM %%s WHERE id >= ? ORDER BY id, part");	     

#ifdef	XXX_SLF_API_CHANGE_HACK
    r = prep_select(tablename, curstate->odbc_env, &hstmt, queryfmt,
		    curstate->in_idbuf, curstate->in_idbuflen,
		    curstate->res_idbuf, curstate->res_idbuflen,
		    curstate->part, curstate->partmax,
		    curstate->res_valbuf, curstate->res_valbuflen,update);
#else	/* XXX_SLF_API_CHANGE_HACK */
    r = prep_select(tablename, curstate->odbc_env, &hstmt, queryfmt,
		    curstate->in_idbuf, curstate->in_idbuflen,
		    curstate->res_idbuf, curstate->res_idbuflen,
		    curstate->part, curstate->partmax,
		    curstate->res_valbuf, curstate->res_valbuflen, 1);
#endif	/* XXX_SLF_API_CHANGE_HACK */
    if (r != 0) {
	goto cleanup;
    }

    retcode = SQLExecute(hstmt);
    if (! SQL_SUCCEEDED(retcode)) {
        if (get_native_error(hstmt, &es) < 0) {
            dumpodbclog(curstate->odbc_env->henv, curstate->odbc_env->hdbc, hstmt);
            r = XXX_BUMMER_ERROR;
	} else {
            if (es.nativeerror == CLUSTRA_TIMEOUT ||
		es.nativeerror == CLUSTRA_UPGRADE_LOCK) {
		r = ODBC_TIMEOUT;
            } else {
                print_error(es);
                r = XXX_BUMMER_ERROR;
	    }
	}
	goto cleanup;
    }

    r = do_multipart_fetch(curstate->odbc_env, hstmt, NULL, 0, &resid,&residlen,
			   curstate->res_idbuf, curstate->res_idbuflen,
			   curstate->part, curstate->partmax,
			   curstate->res_valbuf, curstate->res_valbuflen);
    if (r == 0) {
	if (*curstate->res_idbuflen == key_len &&
	    memcmp(curstate->res_idbuf, key, key_len) == 0) {
	    /* Set the value results */
	    *val = resid;
	    *val_len = residlen;
	    /* This statement handle is worth keeping for later use */
	    curstate->cstmt = hstmt;
	} else {
            if (r != ODBC_TIMEOUT)
		r = DB_NOTFOUND;
	}
    }

    /* : cleanup and return */
 cleanup:
    if (r != 0 && hstmt != 0) SQLFreeStmt(hstmt, SQL_DROP);
    return (r);
}

/* END FUNCTION: msdb__cursor_get_exact */


/*  */

/* FUNCTION: msdb__cursor_get_closest */

/* SYNOPSIS
 * Outputed values are nul terminated.
 * END SYNOPSIS */

/* NOTES
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__cursor_get_closest (		/* D: set cursor to key that closest matches pattern */
/* PARAMETERS */
  DBC *cursor,				/* I: cursor pointer */
  void *pattern,			/* I: partial key */
  unsigned int pattern_len,		/* I: length of pattern */
  void **key,				/* O: key */
  unsigned int *key_len,		/* O: length of key data */
  void **val,				/* O: value */
#ifdef	XXX_SLF_API_CHANGE_HACK
  unsigned int *val_len,		/* O: length of val data */
  int update                            /* I: update cursor? */
#else	/* XXX_SLF_API_CHANGE_HACK */
  unsigned int *val_len			/* O: length of val data */
#endif	/* XXX_SLF_API_CHANGE_HACK */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;				/* generic return code */
    ODBC_CURSOR_t *curstate;
    char *tablename;
    SQLHSTMT hstmt;
    SQLRETURN retcode;
    char *resid, *resval;
    int residlen, resvallen;
    char queryfmt[ODBC_MAXQUERYLEN];
    odbc_error_s es;
/* END VARIABLES */

/* : sanity checks */
    assert (cursor != NULL);
    assert (key != NULL);
    assert (key_len != 0);
    assert (val != NULL);
    assert (val_len != NULL);
    curstate = (ODBC_CURSOR_t *) cursor;
    /* XXXYYYZZZ not required for reading, eh? assert(odbc_env->txn_p); */
    if (curstate->ctable == NULL) {
	fprintf(stderr, "XXXYYYXXX msdb__cursor_get_closest called without prior open\n");
	syslog(LOG_CRIT, "XXXYYYXXX msdb__cursor_get_closest called without prior open");
	return 1;
    }
    if (curstate->cstmt != 0) {
	fprintf(stderr, "XXXYYYXXX msdb__cursor_get_closest called with previous active cursor\n");
	syslog(LOG_CRIT, "XXXYYYXXX msdb__cursor_get_closest called with previous active cursor");
	return 1;
    }

/* : initialization */
    hstmt = 0;
    resid = NULL;
    residlen = 0;
    resval = NULL;
    resvallen = 0;

    tablename = (char *) curstate->ctable;
fprintf(stderr, "TRACE: msdb__cursor_get_closest: thread 0x%lx: table = %s ... pattern = 0x%lx for %d, cursor = 0x%lx\n", (unsigned long) pthread_self(), tablename, (unsigned long) pattern, pattern_len, (unsigned long) cursor);
    snprintf(curstate->ccurname, ODBC_MAXLEN_CNAME, "mycursor0");
    snprintf(queryfmt, ODBC_MAXQUERYLEN,
	     "SELECT * FROM %%s WHERE id >= ? ORDER BY id, part");	     

#ifdef	XXX_SLF_API_CHANGE_HACK
    r = prep_select(tablename, curstate->odbc_env, &hstmt, queryfmt,
		    pattern, &pattern_len,
		    curstate->res_idbuf, curstate->res_idbuflen,
		    curstate->part, curstate->partmax,
		    curstate->res_valbuf, curstate->res_valbuflen,update);
#else	/* XXX_SLF_API_CHANGE_HACK */
    r = prep_select(tablename, curstate->odbc_env, &hstmt, queryfmt,
		    pattern, &pattern_len,
		    curstate->res_idbuf, curstate->res_idbuflen,
		    curstate->part, curstate->partmax,
		    curstate->res_valbuf, curstate->res_valbuflen, 1);
#endif	/* XXX_SLF_API_CHANGE_HACK */
    if (r != 0) {
	goto cleanup;
    }

    retcode = SQLExecute(hstmt);
    if (! SQL_SUCCEEDED(retcode)) {
        if (get_native_error(hstmt, &es) < 0) {
            dumpodbclog(curstate->odbc_env->henv, curstate->odbc_env->hdbc, hstmt);
            r = XXX_BUMMER_ERROR;
	} else {
            if (es.nativeerror == CLUSTRA_TIMEOUT ||
		es.nativeerror==CLUSTRA_UPGRADE_LOCK) {
		r = ODBC_TIMEOUT;
            } else {
                print_error(es);
                r = XXX_BUMMER_ERROR;
	    }
	}
	goto cleanup;
    }

    r = do_multipart_fetch(curstate->odbc_env, hstmt, &resid, &residlen,
			   &resval, &resvallen,
			   curstate->res_idbuf, curstate->res_idbuflen,
			   curstate->part, curstate->partmax,
			   curstate->res_valbuf, curstate->res_valbuflen);
    if (r == 0) {
	/* Set the value results */
	*key = resid;
	*key_len = residlen;
	*val = resval;
	*val_len = resvallen;
	/* This statement handle is worth keeping for later use */
	curstate->cstmt = hstmt;
    }

/* : cleanup and return */
 cleanup:
    if (r != 0 && hstmt != 0) SQLFreeStmt(hstmt, SQL_DROP);
    return (r);
}

/* END FUNCTION: msdb__cursor_get_closest */

/*  */

/* FUNCTION: msdb__cursor_get_first */

/* SYNOPSIS
 * Outputed values are nul terminated.
 * END SYNOPSIS */

/* NOTES
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__cursor_get_first (		/* D: set cursor to first record */
/* PARAMETERS */
  DBC *cursor,				/* I: cursor pointer */
  void **key,				/* O: key */
  unsigned int *key_len,		/* O: length of key data */
  void **val,				/* O: value */
#ifdef	XXX_SLF_API_CHANGE_HACK
  unsigned int *val_len,		/* O: length of val data */
  int update                            /* I: update cursor? */
#else	/* XXX_SLF_API_CHANGE_HACK */
  unsigned int *val_len			/* O: length of val data */
#endif	/* XXX_SLF_API_CHANGE_HACK */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;				/* generic return code */
    ODBC_CURSOR_t *curstate;
    char *tablename;
    SQLHSTMT hstmt;
    SQLRETURN retcode;
    char *resid, *resval;
    int residlen, resvallen;
    char queryfmt[ODBC_MAXQUERYLEN];
    odbc_error_s es;
/* END VARIABLES */

/* : sanity checks */
    assert (cursor != NULL);
    assert (key != NULL);
    assert (key_len != 0);
    assert (val != NULL);
    assert (val_len != NULL);
    curstate = (ODBC_CURSOR_t *) cursor;
    /* XXXYYYZZZ not required for reading, eh? assert(odbc_env->txn_p); */
    if (curstate->ctable == NULL) {
	fprintf(stderr, "XXXYYYXXX msdb__cursor_get_first called without prior open\n");
	syslog(LOG_CRIT, "XXXYYYXXX msdb__cursor_get_first called without prior open");
	return 1;
    }
    if (curstate->cstmt != 0) {
	fprintf(stderr, "XXXYYYXXX msdb__cursor_get_first called with previous active cursor\n");
	syslog(LOG_CRIT, "XXXYYYXXX msdb__cursor_get_first called with previous active cursor");
	return 1;
    }

/* : initialization */
    hstmt = 0;
    resid = NULL;
    residlen = 0;
    resval = NULL;
    resvallen = 0;

    tablename = (char *) curstate->ctable;
fprintf(stderr, "TRACE: msdb__cursor_get_first: thread 0x%lx: table = %s, cursor = 0x%lx\n", (unsigned long) pthread_self(), tablename, (unsigned long) cursor);
    snprintf(curstate->ccurname, ODBC_MAXLEN_CNAME, "mycursor0");
    snprintf(queryfmt, ODBC_MAXQUERYLEN,
	     "SELECT * FROM %%s ORDER BY id, part");	     

#ifdef	XXX_SLF_API_CHANGE_HACK
    r = prep_select(tablename, curstate->odbc_env, &hstmt, queryfmt, NULL, 0,
		    curstate->res_idbuf, curstate->res_idbuflen,
		    curstate->part, curstate->partmax,
		    curstate->res_valbuf, curstate->res_valbuflen,update);
#else	/* XXX_SLF_API_CHANGE_HACK */
    r = prep_select(tablename, curstate->odbc_env, &hstmt, queryfmt, NULL, 0,
		    curstate->res_idbuf, curstate->res_idbuflen,
		    curstate->part, curstate->partmax,
		    curstate->res_valbuf, curstate->res_valbuflen, 1);
#endif	/* XXX_SLF_API_CHANGE_HACK */
    if (r != 0) {
	goto cleanup;
    }

    retcode = SQLExecute(hstmt);
    if (! SQL_SUCCEEDED(retcode)) {
        if (get_native_error(hstmt, &es) < 0) {
            dumpodbclog(curstate->odbc_env->henv, curstate->odbc_env->hdbc, hstmt);
            r = XXX_BUMMER_ERROR;
	} else {
            if (es.nativeerror == CLUSTRA_TIMEOUT ||
		es.nativeerror == CLUSTRA_UPGRADE_LOCK) {
		r = ODBC_TIMEOUT;
            } else {
                print_error(es);
                r = XXX_BUMMER_ERROR;
	    }
	}
	goto cleanup;
    }

    r = do_multipart_fetch(curstate->odbc_env, hstmt, &resid, &residlen,
			   &resval, &resvallen,
			   curstate->res_idbuf, curstate->res_idbuflen,
			   curstate->part, curstate->partmax,
			   curstate->res_valbuf, curstate->res_valbuflen);
    if (r == 0) {
	/* Set the value results */
	*key = resid;
	*key_len = residlen;
	*val = resval;
	*val_len = resvallen;
	/* This statement handle is worth keeping for later use */
	curstate->cstmt = hstmt;
    }

/* : cleanup and return */
 cleanup:
    if (r != 0 && hstmt != 0) SQLFreeStmt(hstmt, SQL_DROP);
    return (r);
}

/* END FUNCTION: msdb__cursor_get_first */


/*  */

/* FUNCTION: msdb__cursor_get_next */

/* SYNOPSIS
 * END SYNOPSIS */

/* NOTES
 * Outputed values are nul terminated.
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__cursor_get_next (			/* D: get next record */
/* PARAMETERS */
  DBC *cursor,				/* I: cursor pointer */
  void **key,				/* O: key */
  unsigned int *key_len,		/* O: length of key data */
  void **val,				/* O: value */
  unsigned int *val_len			/* O: length of val data */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;				/* generic return code */
    ODBC_CURSOR_t *curstate;
    char *resid, *resval;
    int residlen, resvallen;
/* END VARIABLES */

/* : sanity checks */
    if(cursor==NULL)
      fprintf(stderr,"\ncursor == NULL(%d)\n",pthread_self());
    assert (cursor != NULL);
    assert (key != NULL);
    assert (key_len != 0);
    assert (val != NULL);
    assert (val_len != 0);
    curstate = (ODBC_CURSOR_t *) cursor;
    /* XXXYYYZZZ not required for reading, eh? assert(odbc_env->txn_p); */
    if (curstate->cstmt == 0)
	fprintf(stderr,"curstate->cstmt == 0 (%d)\n",pthread_self());
    assert(curstate->cstmt != 0);

fprintf(stderr, "TRACE: msdb__cursor_get_next: thread 0x%lx: table = %s, cursor = 0x%lx\n", (unsigned long) pthread_self(), (char *) curstate->ctable, (unsigned long) cursor);
/* : initialization */
    r = do_multipart_fetch(curstate->odbc_env, curstate->cstmt,
			   &resid, &residlen, &resval, &resvallen,
			   curstate->res_idbuf, curstate->res_idbuflen,
			   curstate->part, curstate->partmax,
			   curstate->res_valbuf, curstate->res_valbuflen);

/* : set output values */
    if (r == 0) {
/* : set output values */
/* : DOC - we tack on a null char at end of string for convenience of caller */
	*(resid + residlen) = '\0';
	*key = resid;
	*key_len = residlen;
/* : DOC - we tack on a null char at end of string for convenience of caller */
	*(resval + resvallen) = '\0';
	*val = resval;
	*val_len = resvallen;
    }

/* : cleanup and return */
    goto cleanup;	/* XXXYYYXXX shut up the compiler */
 cleanup:

    return (r);
}

/* END FUNCTION: msdb__cursor_get_next */


/*  */

/* FUNCTION: msdb__cursor_put_current */

/* SYNOPSIS
 * END SYNOPSIS */

/* NOTES
 * Records are implicitly locked by the transaction subsytem when cursor opened with non-null tid.
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__cursor_put_current (		/* D: overwrite current record */
/* PARAMETERS */
  DBC *cursor,				/* I: cursor pointer */
  void *key,				/* I: key */
  unsigned int key_len,			/* I: length of key data */
  void *val,				/* I: value */
  unsigned int val_len			/* I: length of key val */
/* END PARAMETERS */
)
{
  ODBC_CURSOR_t *curstate;

  /* : sanity checks */
  assert (cursor != NULL);
  assert (key != NULL);
  assert ((val != NULL && val_len != 0) || (val == NULL && val_len == 0));

  curstate = (ODBC_CURSOR_t *) cursor;

fprintf(stderr, "TRACE: msdb__cursor_put_current: thread 0x%lx: table = %s ... key = 0x%lx for %d, val = 0x%lx for %d, cursor = 0x%lx\n", (unsigned long) pthread_self(), (char *) curstate->ctable, (unsigned long) key, key_len, (unsigned long) val, val_len, (unsigned long) cursor);
  /* : initialization */
    
  return msdb__record_put_flags((DB *) curstate->ctable,
				(DB_TXN *)curstate->odbc_env,
                                key, key_len, val, val_len, 0);
}

/* END FUNCTION: msdb__cursor_put_current */


/*  */

/* FUNCTION: msdb__cursor_put_record */

/* SYNOPSIS
 * END SYNOPSIS */

/* NOTES
 * Records are implicitly locked by the transaction subsytem when cursor opened with non-null tid.
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__cursor_put_record (		/* D: add a record */
/* PARAMETERS */
  DBC *cursor,				/* I: cursor pointer */
  void *key,				/* I: key */
  unsigned int key_len,			/* I: length of key data */
  void *val,				/* I: value */
  unsigned int val_len			/* I: length of key val */
/* END PARAMETERS */
)
{
  /* XXXYYYZZZ We're assuming msdb__cursor_put_current is the same */
fprintf(stderr, "TRACE: msdb__cursor_put_record: thread 0x%lx: key = 0x%lx for %d, val = 0x%lx for %d, cursor = 0x%lx\n", (unsigned long) pthread_self(), (unsigned long) key, key_len, (unsigned long) val, val_len, (unsigned long) cursor);
  return msdb__cursor_put_current(cursor, key, key_len, val, val_len);
}

/* END FUNCTION: msdb__cursor_put_record */


/*  */

/* FUNCTION: msdb__cursor_del_current */

/* SYNOPSIS
 * END SYNOPSIS */

/* NOTES
 * Records are implicitly locked by the transaction subsytem when cursor opened with non-null tid.
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__cursor_del_current (		/* D: delete current record */
/* PARAMETERS */
  DBC *cursor				/* I: cursor pointer */
/* END PARAMETERS */
)
{
  ODBC_CURSOR_t *curstate;

  /* : sanity checks */
  assert (cursor != NULL);

  /* : initialization */
  curstate = (ODBC_CURSOR_t *) cursor;
    
  /*
  ** XXXYYYZZZ We're assuming that all the cursor ops are using
  ** prep_select() and thus the record read via the cursor will have
  ** bound the key column to curstate->res_idbuf{,len}.  Warranty void
  ** if this label removed.
  */
fprintf(stderr, "TRACE: msdb__cursor_del_current: thread 0x%lx: cursor = 0x%lx\n", (unsigned long) pthread_self(), (unsigned long) cursor);
  return msdb__record_del((DB *) curstate->ctable, (DB_TXN *)curstate->odbc_env,
                          curstate->res_idbuf, *curstate->res_idbuflen);
}

/* END FUNCTION: msdb__cursor_del_current */

/*  */

/* FUNCTION: msdb__transaction_begin */

/* SYNOPSIS
 * Begin a transaction.
 * END SYNOPSIS */

/* NOTES
 * END NOTES */

int                                     /* R: 0 if ok, else errno */
msdb__transaction_begin (               /* D: begin a trasaction */
/* PARAMETERS */
  DB_ENV *env,                          /* I: transaction manager */
  DB_TXN *pid,                          /* I: parent transaction id */
  DB_TXN **tid                          /* O: transaction id */
/* END PARAMETERS */
)
{
  /* VARIABLES */
  ODBC_ENV_t *odbc_env;
  SQLRETURN retcode;
  /* END VARIABLES */

  /* : sanity checks */
  assert (env != NULL);
  /* pid is unused in the ODBC scheme of things.... */
  assert (tid != NULL);

  /* : initialization */
  odbc_env = get_per_thread_odbc_env();
  assert(odbc_env != NULL);
  assert((! odbc_env->txn_p) || (odbc_env->txn_p && pid != NULL));
  *tid = (DB_TXN *) odbc_env;

fprintf(stderr, "TRACE: msdb__transaction_begin: thread 0x%lx: pid = 0x%lx, tid = 0x%lx\n", (unsigned long) pthread_self(), (unsigned long) pid, (unsigned long) *tid);
  if (pid == NULL) {
      /* Our previous assert() makes these initializations safe */
      odbc_env->txn_p = 1;
      odbc_env->ctxn_p = 0;
      if (odbc_env->autoc_p) {	/* Autocommit mode on?  */
	retcode = SQLSetConnectOption(odbc_env->hdbc, SQL_AUTOCOMMIT,
                                  SQL_AUTOCOMMIT_OFF);
	assert(retcode == SQL_SUCCESS);
	odbc_env->autoc_p = 0;
fprintf(stderr, "TRACE: msdb__transaction_begin: thread 0x%lx: pid = 0x%lx, tid = 0x%lx, AUTOCOMMIT now off\n", (unsigned long) pthread_self(), (unsigned long) pid, (unsigned long) *tid);
      }
  } else {
      if (odbc_env->ctxn_p) {
          /* We only support child txns 1 level deep */
          fprintf(stderr, "XXXYYYXXX We only support child txns 1 level deep\n");
          return 1;
      }
      odbc_env->ctxn_p = 1;
  }

  /* : cleanup and return */
  return (0);
}

/* END FUNCTION: msdb__transaction_begin */


/*  */

/* FUNCTION: msdb__transaction_end */

/* SYNOPSIS
 * End a transaction.
 * END SYNOPSIS */

/* NOTES
 * END NOTES */

int                                     /* R: 0 if ok, else errno */
msdb__transaction_end (                 /* D: end a transaction */
/* PARAMETERS */
  DB_TXN *tid                           /* I: transaction id */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;                              /* generic return code */
    ODBC_ENV_t *odbc_env;
    SQLRETURN retcode;
/* END VARIABLES */

/* : sanity checks */
    assert (tid != NULL);

/* : initialization */
    odbc_env = get_per_thread_odbc_env();
    
    assert(odbc_env != NULL);
    assert(odbc_env->txn_p);

fprintf(stderr, "TRACE: msdb__transaction_end: thread 0x%lx: tid = 0x%lx\n", (unsigned long) pthread_self(), (unsigned long) tid);
    if (odbc_env->ctxn_p) {
	/* Child transactions are no-ops */
	odbc_env->ctxn_p = 0;
	return 0;
    }
    r = msdb__transaction_end_2 (odbc_env, SQL_COMMIT);
    assert(odbc_env->autoc_p == 0);	/* Something very wrong if fail */
    retcode = SQLSetConnectOption(odbc_env->hdbc, SQL_AUTOCOMMIT,
                                  SQL_AUTOCOMMIT_ON);    
    assert(retcode == SQL_SUCCESS);
    odbc_env->autoc_p = 1;
fprintf(stderr, "TRACE: msdb__transaction_end: thread 0x%lx: tid = 0x%lx, AUTOCOMMIT now on\n", (unsigned long) pthread_self(), (unsigned long) tid);

/* : cleanup and return */
    return (r);
}

/* END FUNCTION: msdb__transaction_end */


/*  */

/* FUNCTION: msdb__transaction_abort */

/* SYNOPSIS
 * Abort a transaction.
 * END SYNOPSIS */

/* NOTES
 * END NOTES */

int                                     /* R: 0 if ok, else errno */
msdb__transaction_abort (                 /* D: abort a transaction */
/* PARAMETERS */
  DB_TXN *tid                           /* I: transaction id */
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;                              /* generic return code */
    ODBC_ENV_t *odbc_env;
/* END VARIABLES */

/* : sanity checks */
    assert (tid != NULL);

/* : initialization */
    odbc_env = get_per_thread_odbc_env();
    assert(odbc_env != NULL);
fprintf(stderr, "TRACE: msdb__transaction_end: thread 0x%lx: tid = 0x%lx, odbc_env->ctxn_p = %d\n", (unsigned long) pthread_self(), (unsigned long) tid, odbc_env->ctxn_p);
    if (odbc_env->ctxn_p) {
	/* Child transactions are no-ops */
	odbc_env->ctxn_p = 0;
	return 0;
    }
    r = msdb__transaction_end_2 (odbc_env, SQL_ROLLBACK);

/* : cleanup and return */
    return (r);
}

/* END FUNCTION: msdb__transaction_abort */

/*  */

/* FUNCTION: msdb__transaction_end_2 (PRIVATE) */

int                                     /* R: 0 if ok, else errno */
msdb__transaction_end_2 (               /* D: end a transaction, backend func  */
/* PARAMETERS */
  ODBC_ENV_t *odbc_env,                  /* I: ODBC environment  */
  UWORD mode
/* END PARAMETERS */
)
{
/* VARIABLES */
    int r;                              /* generic return code */
    SQLRETURN retcode;
/* END VARIABLES */

/* : sanity checks */

/* : initialization */
    retcode = SQLTransact(odbc_env->henv, odbc_env->hdbc, mode);
fprintf(stderr, "TRACE: msdb__transaction_end_2: thread 0x%lx: odbc_env = 0x%lx, retcode = %d\n", (unsigned long) pthread_self(), (unsigned long) odbc_env, retcode);
    if (retcode == SQL_SUCCESS) {
	r = 0;
    } else if (retcode == SQL_SUCCESS_WITH_INFO) {
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, 0);
	r = 0;
    } else {
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, 0);
	r = XXX_BUMMER_ERROR;
    }
    odbc_env->txn_p = 0;
    
/* : cleanup and return */
    return (r);
}

/* END FUNCTION: msdb__transaction_end_2 */

/*
** encode_binary() -- Encode a binary data buffer so that it's NUL-free.
**	I'd really prefer to use SQLBindParameter() with a SQL_BINARY type,
**	but it doesn't work (sends '' to the PostgreSQL server instead),
**	and I don't know if it's unixODBC's fault or the odbcpgsql driver's
**	fault.
**
**	For each NUL byte we find, we encode it as two bytes: ESC_VAL
**	and ESC_NUL.  The ESC_VAL byte itself is encoded as the two bytes
**	ESC_VAL and ESC_ESC.  This scheme works well for PostgreSQL:
**	the ODBC driver seems to handle ASCII values 1-255 just fine.
**	PostgreSQL's binary type, BYTEA, is unbounded by default.
**	Clustra's max BINARY size is apparently 8K (the standard max
**	column width), and apparently BINARY columns are stored as fixed-
**	width entities, like CHAR, as opposed to space-saving VARCHAR.
**
**	PostgreSQL note: it looks like \n characters inside a char string
**	are bad news.  The identifiers table defaults for minimum udid
**	and uuid is ((unsigned long long) 10).  This ends up being stored
**	as 15 bytes in 'val'.  "SELECT *, length(val) from identifiers" shows
**	that 15 bytes are stored.  But when retrieving that row, the 
**	raw SQLFetch() result returns 16 bytes AAAZZZAAAZZZ!
**
**	As long as Clustra's
**	ODBC driver (and whatever driver manager we end up using) can
**	also handle ASCII 1-255 without problems, this encoding scheme
**	will be fine.  It's got the penalty of copying the data an 
**	extra time, but that seems preferable to the penalty of storing
**	a lot of right-padded NUL bytes on disk.  {sigh}
**
**	ASSUMPTIONS: dst is allocated by the caller and is twice the size
**		that src points to.
*/

#define	ESC_VAL	255
#define ESC_NUL 1
#define ESC_ESC 2
#define ESC_LF  3
#define ESC_CR  4

static int
encode_binary(const unsigned char *src, int srclen, unsigned char *dst)
{
    int dstlen = 0;

    for ( ; srclen > 0; srclen--, src++, dst++, dstlen++) {
	switch (*src) {
	case 0:		/* NUL */
	    *dst++ = ESC_VAL;
	    dstlen++;
	    *dst   = ESC_NUL;
	    break;
	case 10:	/* LF */
	    *dst++ = ESC_VAL;
	    dstlen++;
	    *dst   = ESC_LF;
	    break;
	case 13:	/* CR */
	    *dst++ = ESC_VAL;
	    dstlen++;
	    *dst   = ESC_CR;
	    break;
	case ESC_VAL:
	    *dst++ = ESC_VAL;
	    dstlen++;
	    *dst   = ESC_ESC;
	    break;
	default:
	    *dst = *src;
	    break;
	}
    }
    return dstlen;
}

/*
** NOTE: we allocate at least 1 byte more than necessary because the
**       caller is probably going to want to NULL-terminate the value
**       "for the convenience of the caller".
**
** The 4th arg, dstbaseprealloced, should be NULL if decode_binary()
** is expected to allocate a buffer to store the result in.  If not
** NULL, that buffer will be used.  In either case, the 3rd arg will
** be used to communicate where the result buffer is located.
*/

static int
decode_binary(const unsigned char *src, int srclen, unsigned char **dstp,
	      unsigned char *dstbaseprealloced)
{
    char	*dstbase, *dst;
    int 	dstlen = 0;

    if (dstbaseprealloced == NULL) {
	if ((dstbase = (char *) fs_get(srclen + 1)) == NULL) {
	    /* XXX *dstp = NULL; */
	    return -1;
	}
    } else {
	dstbase = dstbaseprealloced;
    }
    for (dst = dstbase; srclen > 0; srclen--, src++, dst++, dstlen++) {
	switch  (*src) {
	case ESC_VAL:
	    if (*(src + 1) == ESC_NUL)
		*dst = 0;
	    else if (*(src + 1) == ESC_LF)
		*dst = 10;
	    else if (*(src + 1) == ESC_CR)
		*dst = 13;
	    else
		*dst = ESC_VAL;
	    src++;
	    srclen--;
	    break;
	default:
	    *dst = *src;
	    break;
	}
    }
    *dstp = dstbase;
    return dstlen;
}

static int
prep_select(char *tablename, ODBC_ENV_t *odbc_env, SQLHSTMT *hstmtp,
	    char *queryfmt,
	    char *in_id, int *in_idlen, char *out_id, int *out_idlen,
	    int *out_part, int *out_partmax, char *out_val, int *out_vallen,
	    int update)
{
  int	r = 0;
  char	query[XXX_MAXQUERYLEN];
  SQLRETURN	retcode;
  SQLHSTMT	hstmt;
  
  retcode = SQLAllocStmt(odbc_env->hdbc, &hstmt);
  if (retcode != SQL_SUCCESS) {
      fprintf(stderr,"XXXYYYXXX prep_select allocstmt ret=%d\n",retcode);
      dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
      r = XXX_BUMMER_ERROR;
      goto cleanup;
  }
  if (update) {
      retcode=SQLSetStmtOption(hstmt,SQL_CONCURRENCY,SQL_CONCUR_LOCK);
      if (retcode != SQL_SUCCESS) {
          fprintf(stderr,"XXXYYYXXX setting concurrency to 'lock' failed, ret=%d\n",retcode);
          dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
          r = XXX_BUMMER_ERROR;
          goto cleanup;
      }
  }
  snprintf(query, XXX_MAXQUERYLEN, queryfmt, tablename);
  retcode = SQLPrepare(hstmt, query, SQL_NTS);
  if (retcode != SQL_SUCCESS) {
      fprintf(stderr,"XXXYYYXXX prep_select prepare ret=%d\n",retcode);
      dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
      r = XXX_BUMMER_ERROR;
      goto cleanup;
  }

  /* Bind input parameters */
  if (in_id != NULL) {
      retcode = SQLBindParameter(hstmt, 1, SQL_PARAM_INPUT, SQL_C_CHAR,
                                 SQL_CHAR, *in_idlen, 0, (SQLPOINTER) in_id,
                                 *in_idlen, (SQLINTEGER *) in_idlen);
      if (retcode != SQL_SUCCESS) {
          fprintf(stderr, "XXXYYYXXX SQLBindParameter #1 failed ret=%d\n",retcode);
          dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
          r = XXX_BUMMER_ERROR;
          goto cleanup;
      }
  }

    /* Bind output parameters */
    retcode = SQLBindCol(hstmt, 1, SQL_C_CHAR, (SQLPOINTER) out_id,
			 ODBC_MAXLEN_ID+1, (SQLINTEGER *) out_idlen);
    if (retcode != SQL_SUCCESS) {
	fprintf(stderr, "XXXYYYXXX SQLBindCol #1 failed ret=%d\n",retcode);
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    retcode = SQLBindCol(hstmt, 2, SQL_C_SLONG, (SQLPOINTER) out_part,
			 sizeof(int), NULL);
    if (retcode != SQL_SUCCESS) {
	fprintf(stderr, "XXXYYYXXX SQLBindCol #2 failed ret=%d\n",retcode);
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    retcode = SQLBindCol(hstmt, 3, SQL_C_SLONG, (SQLPOINTER) out_partmax,
			 sizeof(int), NULL);
    if (retcode != SQL_SUCCESS) {
	fprintf(stderr, "XXXYYYXXX SQLBindCol #3 failed ret=%d\n",retcode);
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }
    retcode = SQLBindCol(hstmt, 4, SQL_C_CHAR, (SQLPOINTER) out_val,
			 ODBC_MAXLEN_VAL+1, (SQLINTEGER *) out_vallen);
    if (retcode != SQL_SUCCESS) {
	fprintf(stderr, "XXXYYYXXX SQLBindCol #4 failed ret=%d\n",retcode);
	dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
	r = XXX_BUMMER_ERROR;
	goto cleanup;
    }

    /* If we're here, our preparations were successful */
    *hstmtp = hstmt;
    
 cleanup:
    if (r != 0)
      SQLFreeStmt(hstmt, SQL_DROP);
    return r;
}

static int
do_multipart_fetch(ODBC_ENV_t *odbc_env, SQLHSTMT hstmt, char **dstid,
		   int *dstidlen, char **dstval, int *dstvallen,
		   char *idbuf, int *idbuflen,
		   int *partp, int *partmaxp,
		   char *valbuf, int *valbuflen)
{
    int r = 0;
    SQLRETURN retcode;
    unsigned char	*decbufp;
    int decbufplen;
    int real_part, real_partmax;
    char real_id[ODBC_MAXLEN_ID+1];
    int real_idlen;
    char *bigresult = NULL, *bigresultp = NULL;
    int bigresultlen = 0, chunksize;
    pthread_t self= pthread_self();	/* XXXYYYZZZ debugging */
    odbc_error_s es;
  
    *partp = 0;
    *partmaxp = 0;

  /*
   * if this block is executed more than once, we are trying to read a
   * record that is in the process of beeing inserted. This record can
   * safely be droped since it was just inserted (this thread could not
   * posibly know that it was supposed to be here).
   * This is because of the phantom read problem, will not be fixed before
   * Clustra supports range locks (isolation level serializable).
   *
   * If this happens a write cursor have gone past a read cursor
   */
    do {
	if (*partp != 0)
	    fprintf(stderr,"\n\n\n\n****   do_multipart_fetch(%d) part=%d   ****\n\n\n\n",self,*partp);
     
	retcode = SQLFetch(hstmt);

	if (retcode == SQL_NO_DATA) {
	    r = DB_NOTFOUND;
	    goto cleanup;
	} else if (retcode != SQL_SUCCESS) {
	    if (get_native_error(hstmt, &es) < 0) {
		dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
		fprintf(stderr, "XXXYYYXXX retcode = %d\n", retcode);
		r = XXX_BUMMER_ERROR;
	    } else {
		if (es.nativeerror == CLUSTRA_TIMEOUT ||
		    es.nativeerror == CLUSTRA_UPGRADE_LOCK) {
		    r = ODBC_TIMEOUT;
		} else {
		    print_error(es);
		    r = XXX_BUMMER_ERROR;
		}
	    }
	    goto cleanup;
	}
    } while(*partp!=1);
    
    assert(*partp == 1);
 
    decbufplen = decode_binary(valbuf, *valbuflen, &decbufp, NULL);
    if (*partp == 1 && *partmaxp == 1) {
	*dstval = decbufp;
	*dstvallen = decbufplen;
	if (dstid != NULL) {
	    decbufplen = decode_binary(idbuf, *idbuflen, &decbufp, NULL);
	    *dstid = decbufp;
	    *dstidlen = decbufplen;
	}
	r = 0;
	goto cleanup;
    } else {
	/* Here's where we earn our bread & butter */
	assert(*partmaxp > 1);
	/* Keep track of this 1st row's info, just to be paranoid */
	real_part = *partp;
	real_partmax = *partmaxp;
	memcpy(real_id, idbuf, *idbuflen);
	real_idlen = *idbuflen;
	
     /* Allocate a dynamic buffer to store all the chunks */
	if ((bigresult =
	     (char *) fs_get((*partmaxp * ODBC_MAXLEN_VAL) + 1)) == NULL) {
	    r = errno;
	    goto cleanup;
	}
	/* Copy the 1st chunk (in the static buffer) to the dynamic one */
	bigresultp = bigresult;
	memcpy(bigresultp, decbufp, decbufplen);
	bigresultp += decbufplen;
	bigresultlen = decbufplen;
	/* Put the rest of the chunks in the dynamic buffer */
	for (real_part = 2 ; real_part <= real_partmax; real_part++) {
	    /* Rebind col #4 to put new chunk inside our dynamic buffer */
	    retcode = SQLBindCol(hstmt, 4, SQL_C_CHAR,
				 (SQLPOINTER) bigresultp, ODBC_MAXLEN_VAL+1,
				 (SQLINTEGER *) &chunksize);
	    if (retcode != SQL_SUCCESS) {
		fprintf(stderr, "XXXYYYXXX SQLBindCol #4 (rebind) failed: %d\n", retcode);
		dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
		r = XXX_BUMMER_ERROR;
		goto cleanup;
	    }
	    retcode = SQLFetch(hstmt);
	    if (retcode == SQL_SUCCESS) {
		/* Paranoid error checking here: detect missing rows */
		if(*partp!=real_part)
		    fprintf(stderr,"\n part = %d \n",*partp);
		assert(*partp == real_part);
		assert(*partmaxp == real_partmax);
		assert(*idbuflen == real_idlen);
		assert(memcmp(idbuf, real_id, real_idlen) == 0);
	    } else {
		/*
             ** XXXYYYXXX This is a pretty serious problem: something
             ** really icky happened to the server, or our database is
             ** corrupt, i.e. missing rows that ought to be here.
             ** So, what should be done?
             */
		if (get_native_error(hstmt, &es) < 0) {
		    fprintf(stderr, "XXXYYYXXX Gack!  Get chunk failed, retcode = %d\n", retcode);
		    *(idbuf + *idbuflen)=0;
		    fprintf(stderr,"\nid=%s, part=%d, max=%d\n",idbuf,*partp,*partmaxp);
		    dumpodbclog(odbc_env->henv, odbc_env->hdbc, hstmt);
		    r = DB_NOTFOUND;
		} else {
		    if (es.nativeerror == CLUSTRA_TIMEOUT || 
			es.nativeerror == CLUSTRA_UPGRADE_LOCK) {
			r = ODBC_TIMEOUT;
		    } else {
			print_error(es);
			r = DB_NOTFOUND;
		    }
		    goto cleanup;
		}
	    }
	    bigresultp += chunksize;
	    bigresultlen += chunksize;
	}
	decbufplen = decode_binary(bigresult, bigresultlen, &decbufp, NULL);
	*dstval = decbufp;
	*dstvallen = decbufplen;
	if (dstid != NULL) {
	    decbufplen = decode_binary(idbuf, *idbuflen, &decbufp, NULL);
	    *dstid = decbufp;
	    *dstidlen = decbufplen;
	}
	r = 0;
	goto cleanup;
    }

 cleanup:
    /* decode_binary() made yet another buffer for us, don't need bigresult */
    if (bigresult != NULL) {
	fs_give((void **) &bigresult);
	/*
	** XXX Whee!  This was a tough bug to find!  If we read a multi-chunk
	** value, we'd used SQLBindCol to put stuff into a dynamic buffer
	** (bigresult) allocated by this function.
	** If we don't put bind it back to valbuf, we will
	** clobber memory bigtime the next time we try a cursor-based
	** fetch.
	*/
	SQLBindCol(hstmt, 4, SQL_C_CHAR, (SQLPOINTER) valbuf,
		   ODBC_MAXLEN_VAL+1, (SQLINTEGER *) valbuflen);
    }
    return r;
}

/*  */

/* FUNCTION: msdb__entries_count */

/* SYNOPSIS
 * Return the number of records in a table.
 * END SYNOPSIS */

/* NOTES
 * This function is very expensive as it may access all pages of the table.
 * This function may return inaccurate data if used in a live environment.
 * END NOTES */

int					/* R: 0 if ok, else errno */
msdb__entries_count (			/* D: return number of records */
/* PARAMETERS */
  DB *db,				/* I: database table */
  long *count				/* U: number of entries */
/* END PARAMETERS */
)
{
/* VARIABLES */
    DBC *cursor;                        /* db cursor */
    void *key;				/* record key */
    void *val;				/* record value */
    unsigned int keylen;		/* record key length */
    unsigned int vallen;		/* record value length */
    int return_value;			/* return value */
    long local_count;			/* local record count */
/* END VARIABLES */

/* : sanity checks */
    assert (db != NULL);
    assert (count != NULL);

/* : initialization */
    key = NULL;
    val = NULL;
    return_value = 0;
    local_count = 0;

/* : make cursor */
    return_value = msdb__cursor_open (db, NULL, &cursor);
    if (return_value != 0) {
	goto cleanup;
    }

/* : get first record in the table */
#ifdef	XXX_SLF_API_CHANGE_HACK
    return_value = msdb__cursor_get_first (cursor, &key, &keylen, &val, &vallen,0);
#else	/* XXX_SLF_API_CHANGE_HACK */
    return_value = msdb__cursor_get_first (cursor, &key, &keylen, &val, &vallen);
#endif	/* XXX_SLF_API_CHANGE_HACK */
    
/* : count all records in the table */
    while (return_value == 0) {
	local_count++;
	fs_give ((void **) &key);
	fs_give ((void **) &val);
        return_value = msdb__cursor_get_next (cursor, &key, &keylen, &val, &vallen);
    }

/* : we are done if we walked off the table */
    if (return_value == DB_NOTFOUND) {
	return_value = 0;
    }

/* : cleanup and return status */
 cleanup:
    fs_give ((void **) &key);
    fs_give ((void **) &val);
    msdb__cursor_close (&cursor);
    *count = local_count;
    return (return_value);
}

/* END FUNCTION: msdb__entries_count */

/* END MODULE: MSDB_Low_Level_Access_Routines */

/* XXXYYYXXX debugging */ 
void dumpodbclog(SQLHENV henv, SQLHDBC hdbc, SQLHSTMT hstmt)
{
    SQLCHAR error[XXX_MAXERRLEN+1];	/* ODBC error string */
    SQLCHAR sqlstate[6];		/* ODBC 5-char error string */
    SDWORD nativeerror;			/* ODBC native error code */
    SWORD errorlen;			/* ODBC error string length */

    if (hstmt) {
	    while (SQLError(henv, hdbc, hstmt, sqlstate, &nativeerror,
			    error, XXX_MAXERRLEN, &errorlen) == SQL_SUCCESS) 
	       {
		  fprintf(stderr,"\nError on HSTMT");
		  fprintf(stderr,"\nSQLSTATE   : %s",sqlstate);
      	          fprintf(stderr,"\nNative Code: %ld",nativeerror);
      	          fprintf(stderr,"\nError Msg  : %s\n",error);
	       }
    }
    if (hdbc) {
	    while (SQLError(henv, hdbc, 0, sqlstate, &nativeerror,
			    error, XXX_MAXERRLEN, &errorlen) == SQL_SUCCESS) 
	       {
	          fprintf(stderr,"\nError on HDBC");
		  fprintf(stderr,"\nSQLSTATE   : %s",sqlstate);
      	          fprintf(stderr,"\nNative Code: %ld",nativeerror);
      	          fprintf(stderr,"\nError Msg  : %s\n",error);
	       }
    }
    if (henv) {
	    while (SQLError(henv, 0, 0, sqlstate, &nativeerror,
			    error, XXX_MAXERRLEN, &errorlen) == SQL_SUCCESS) 
	       {
	          fprintf(stderr,"\nError on HENV");
		  fprintf(stderr,"\nSQLSTATE   : %s",sqlstate);
      	          fprintf(stderr,"\nNative Code: %ld",nativeerror);
      	          fprintf(stderr,"\nError Msg  : %s\n",error);
	       }
    }
}

int get_native_error(SQLHSTMT hstmt,odbc_error_s *error)
{
    if (SQLError(0, 0, hstmt,error->sqlstate, &(error->nativeerror),
		 error->error, ODBC_MAXERRLEN,
		 &(error->errorlen)) != SQL_SUCCESS) {
	fprintf(stderr, "get_native_error(%d): SQLError failed\n",pthread_self());
	return -1;
    }
    return error->nativeerror;
}
  
void print_error(odbc_error_s error)
{
    fprintf(stderr,"\nSQLSTATE   : %s",error.sqlstate);
    fprintf(stderr,"\nNative Code: %ld",error.nativeerror);
    fprintf(stderr,"\nError Msg  : %s\n",error.error);
}
